from OFS.CopySupport import CopyError
from zope.component import queryMultiAdapter
from plone.locking.interfaces import ILockable
from Acquisition import aq_inner
from Acquisition import aq_parent
from castle.cms import tasks
from castle.cms import trash
from castle.cms.browser.utils import Utils
from castle.cms.utils import get_paste_data
from castle.cms.utils import is_max_paste_items
from plone import api
from plone.app.content.browser import actions
from plone.memoize.view import memoize
from plone.uuid.interfaces import IUUID
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone import PloneMessageFactory as _
from Products.CMFPlone.utils import safe_unicode
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.statusmessages.interfaces import IStatusMessage
from z3c.form import button


class ObjectPasteView(actions.ObjectPasteView):

    def paste_async(self, paste_data):
        folder_path = '/'.join(self.context.getPhysicalPath())
        site_path = '/'.join(api.portal.get().getPhysicalPath())
        folder_path = folder_path[len(site_path):]
        tasks.paste_items.delay(
            folder_path, paste_data['op'],
            paste_data['mdatas'])
        return self.do_redirect(
            self.canonical_object_url,
            """You have selected to paste %i items which is a large amount.
This action can take a long time to accomplish. We will email you
when the content is done being moved.""" % paste_data['count'],
            'error'
        )

    def do_action(self):
        paste_data = get_paste_data(self.request)
        if is_max_paste_items(paste_data):
            return self.paste_async(paste_data)
        return super(ObjectPasteView, self).do_action()


class RenameForm(actions.RenameForm):
    template = ViewPageTemplateFile('templates/object_rename.pt')

    @button.buttonAndHandler(_(u'Rename'), name='Rename')
    def handle_rename(self, action):
        lockable = ILockable(self.context)
        if lockable:
            lockable.clear_locks()
        return super(RenameForm, self).handle_rename(self, action)


def is_locked(view):
    locking_view = queryMultiAdapter(
        (view.context, view.request), name='plone_lock_info')

    return locking_view is not None and locking_view.is_locked()


class DeleteConfirmationForm(actions.DeleteConfirmationForm):

    template = ViewPageTemplateFile('templates/delete_confirmation.pt')

    @property
    @memoize
    def items_to_delete(self):
        catalog = getToolByName(self.context, 'portal_catalog')
        results = catalog({
            'path': '/'.join(self.context.getPhysicalPath())
        })
        return len(results)

    def more_info(self):
        if self.items_to_delete > 120:
            IStatusMessage(self.request).add("""
You have selected to delete %i items which is a large amount.
This action can take a long time to accomplish. We will email you
when the content is done being deleted.""" % self.items_to_delete, 'error')
        else:
            return super(DeleteConfirmationForm, self).more_info()

    @property
    def is_locked(self):
        return is_locked(self)

    @button.buttonAndHandler(_(u'Delete'), name='Delete')
    def handle_delete(self, action):
        title = safe_unicode(self.context.Title())
        parent = aq_parent(aq_inner(self.context))

        # has the context object been acquired from a place it should not have
        # been?
        if self.context.aq_chain == self.context.aq_inner.aq_chain:
            if self.items_to_delete > 120:
                tasks.delete_items.delay([IUUID(self.context)])
                IStatusMessage(self.request).add("""
You have selected to delete %i items which is a large amount.
This action can take a long time to accomplish. We will email you
when the content is done being deleted.""" % self.items_to_delete, 'error')
            else:
                if self.is_locked:
                    # force unlock before delete...
                    lockable = ILockable(self.context)
                    lockable.clear_locks()

                parent.manage_delObjects(self.context.getId(), self.request)
                IStatusMessage(self.request).add(
                    _(u'${title} has been deleted.', mapping={u'title': title}))
        else:
            IStatusMessage(self.request).add(
                _(u'"${title}" has already been deleted',
                  mapping={u'title': title})
            )

        self.request.response.redirect(parent.absolute_url())

    @property
    def items_to_trash(self):
        catalog = getToolByName(self.context, 'portal_catalog')
        putils = getToolByName(self.context, 'plone_utils')
        results = catalog({
            'path': '/'.join(self.context.getPhysicalPath()),
            'portal_type': putils.getUserFriendlyTypes(),
        })
        return len(results)

    @button.buttonAndHandler(u'Move to recycle bin', name='Move')
    def handle_move(self, action):
        title = safe_unicode(self.context.Title())
        parent = aq_parent(aq_inner(self.context))

        # has the context object been acquired from a place it should not have
        # been?
        if self.context.aq_chain == self.context.aq_inner.aq_chain:
            trash.object(self.context)
            IStatusMessage(self.request).add(u'%s has been moved to recycle bin.' % title)

        self.request.response.redirect(parent.absolute_url())

    @button.buttonAndHandler(u'Cancel', name='Cancel')
    def handle_cancel(self, action):
        utils = Utils(self.context, self.request)
        target = utils.get_object_url(self.context)
        return self.request.response.redirect(target)


class ObjectCutView(actions.ObjectCutView):

    def do_action(self):
        if is_locked(self):
            # force unlock before delete...
            lockable = ILockable(self.context)
            lockable.clear_locks()

        try:
            self.parent.manage_cutObjects(self.context.getId(), self.request)
        except CopyError:
            return self.do_redirect(self.view_url,
                                    _(u'${title} is not moveable.',
                                        mapping={'title': self.title}))

        return self.do_redirect(
            self.view_url,
            _(u'${title} cut.', mapping={'title': self.title}),
            'info'
        )



from zope.interface import Interface


class ITileView(Interface):

    def __init__(self, tile):
        pass


class ISecureLoginAllowedView(Interface):
    """
    marker interface to describe a view that is allowed to be rendered
    """



# -*- coding: utf-8 -*-
from castle.cms import subscribe
from castle.cms.browser.controlpanel.announcements import SendEmailSubscribersForm
from castle.cms.browser.subscribe import SubscribeForm
from castle.cms.testing import CASTLE_PLONE_INTEGRATION_TESTING
from plone.app.testing import logout
from plone.registry.interfaces import IRegistry
from zope.component import queryUtility

import responses
import transaction
import unittest


class TestSubscribe(unittest.TestCase):

    layer = CASTLE_PLONE_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']
        registry = queryUtility(IRegistry)
        registry['castle.subscriber_categories'] = [u'A', u'B', u'C']

    def test_add_subscriber(self):
        data = subscribe.register('foo@bar.com', {'foo': 'bar'})
        subscribe.confirm('foo@bar.com', data['code'])
        self.assertEquals(len([i for i in subscribe.all()]), 1)

    def test_adds_code(self):
        subscribe.register('foo@bar.com', {'foo': 'bar'})
        item = subscribe.get_subscriber('foo@bar.com')
        self.assertTrue('code' in item)

    def test_get_phone_numbers_empty(self):
        data = subscribe.register('foo@bar.com', {'foo': 'bar'})
        subscribe.confirm('foo@bar.com', data['code'])
        self.assertEquals(len(subscribe.get_phone_numbers()), 0)

    def test_get_phone_numbers(self):
        data = subscribe.register('foo@bar.com', {'foo': 'bar', 'phone_number': '15555555555'})
        subscribe.confirm('foo@bar.com', data['code'])
        subscribe.confirm_phone_number('foo@bar.com', data['code'])
        self.assertEquals(len(subscribe.get_phone_numbers()), 1)

    def test_get_email_address(self):
        data = subscribe.register('foo@bar.com', {'foo': 'bar'})
        subscribe.confirm('foo@bar.com', data['code'])
        self.assertEquals(len(subscribe.get_email_addresses()), 1)

    def test_categories(self):
        data = subscribe.register('foo@bar.com', {'categories': [u'A']})
        subscribe.confirm('foo@bar.com', data['code'])
        user = subscribe.get('foo@bar.com')
        self.assertEquals(user['categories'], [u'A'])
        self.assertFalse(user['categories'] == [])


class TestSubscribeForm(unittest.TestCase):

    layer = CASTLE_PLONE_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']

        registry = queryUtility(IRegistry)
        registry['plone.email_from_address'] = 'foo@bar.com'
        registry['castle.recaptcha_private_key'] = u'foobar'
        registry['castle.recaptcha_public_key'] = u'foobar'
        registry['plone.enable_notification_subscriptions'] = True
        registry['castle.subscriber_categories'] = [u'A', u'B', u'C']

        transaction.commit()

    @responses.activate
    def test_user_subscribes(self):
        responses.add(responses.POST, "https://www.google.com/recaptcha/api/siteverify",
                      body='{"success": true}',
                      content_type="application/json")
        self.request.form.update({
            'form.widgets.name': u'Foobar',
            'form.widgets.email': u'foo@bar.com',
            'g-recaptcha-response': u'foobar',
            'form.buttons.subscribe': 'Subscribe'
        })
        form = SubscribeForm(self.portal, self.request)
        form()
        mailhost = self.portal.MailHost
        self.assertEqual(len(mailhost.messages), 1)
        self.assertEquals(len(subscribe.SubscriptionStorage()._data), 1)

    @responses.activate
    def test_invalid_recaptcha(self):
        logout()
        responses.add(responses.POST, "https://www.google.com/recaptcha/api/siteverify",
                      body='{"success": false}',
                      content_type="application/json")
        self.request.form.update({
            'form.widgets.name': u'Foobar',
            'form.widgets.email': u'foo@bar.com',
            'g-recaptcha-response': u'foobar',
            'form.buttons.subscribe': 'Subscribe'
        })
        form = SubscribeForm(self.portal, self.request)
        form()
        mailhost = self.portal.MailHost
        self.assertEqual(len(mailhost.messages), 0)
        self.assertEquals(len(subscribe.SubscriptionStorage()._data), 0)

    @responses.activate
    def test_user_exists(self):
        subscribe.register('foo@bar.com', {'foo': 'bar'})
        responses.add(responses.POST, "https://www.google.com/recaptcha/api/siteverify",
                      body='{"success": true}',
                      content_type="application/json")
        self.request.form.update({
            'form.widgets.name': u'Foobar',
            'form.widgets.email': u'foo@bar.com',
            'g-recaptcha-response': u'foobar',
            'form.buttons.subscribe': 'Subscribe'
        })
        form = SubscribeForm(self.portal, self.request)
        form()
        mailhost = self.portal.MailHost
        self.assertEqual(len(mailhost.messages), 0)
        self.assertEquals(len(subscribe.SubscriptionStorage()._data), 1)

    @responses.activate
    def test_category_match(self):
        subscribe.register('foo@bar.com', {'categories': [u'A']})
        subscriber = subscribe.get('foo@bar.com')
        subscribe.confirm('foo@bar.com', subscriber['code'])
        responses.add(responses.POST, "https://www.google.com/recaptcha/api/siteverify",
                      body='{"success": true}',
                      content_type="application/json")
        self.request.form.update({
            'form.widgets.subject': u'TEST',
            'form.widgets.send_to_categories': [u'A'],
            'form.widgets.body': 'Words words words',
            'form.widgets.body_plain': 'Words words words',
            'form.buttons.send2': 'Send'
        })
        form = SendEmailSubscribersForm(self.portal, self.request)
        form()
        mailhost = self.portal.MailHost
        self.assertEqual(len(mailhost.messages), 1)
        self.assertEquals(len(subscribe.SubscriptionStorage()._data), 1)

    @responses.activate
    def test_category_mismatch(self):
        subscribe.register('foo@bar.com', {'categories': [u'A']})
        subscriber = subscribe.get('foo@bar.com')
        subscribe.confirm('foo@bar.com', subscriber['code'])

        responses.add(responses.POST, "https://www.google.com/recaptcha/api/siteverify",
                      body='{"success": true}',
                      content_type="application/json")
        self.request.form.update({
            'form.widgets.subject': u'TEST',
            'form.widgets.send_to_categories': [u'B'],
            'form.widgets.body': 'Words words words',
            'form.widgets.body_plain': 'Words words words',
            'form.buttons.send2': 'Send'
        })
        form = SendEmailSubscribersForm(self.portal, self.request)
        form()
        mailhost = self.portal.MailHost
        self.assertEqual(len(mailhost.messages), 0)
        self.assertEquals(len(subscribe.SubscriptionStorage()._data), 1)

    def test_confirm(self):
        subscribe.register('foo@bar.com', {'foo': 'bar'})
        subscriber = subscribe.get_subscriber('foo@bar.com')
        self.request.form.update({
            'confirmed_email': u'foo@bar.com',
            'confirmed_code': subscriber['code']
        })
        form = SubscribeForm(self.portal, self.request)
        form()
        self.assertTrue(form.subscribed)
        mailhost = self.portal.MailHost
        self.assertEqual(len(mailhost.messages), 0)
        self.assertEquals(len([i for i in subscribe.all()]), 1)
        subscriber = subscribe.get_subscriber('foo@bar.com')
        self.assertTrue(subscriber['confirmed'])



# Fork of tweepy twitter stream handling
from plone import api
from plone.registry.interfaces import IRegistry
from requests.exceptions import Timeout
from requests_oauthlib import OAuth1
from requests_oauthlib import OAuth1Session
from threading import Thread
from time import sleep
from zope.component import getUtility
from zope.globalrequest import getRequest

import re
import requests
import six
import ssl


request_token_url = 'https://api.twitter.com/oauth/request_token'
authorization_url = 'https://api.twitter.com/oauth/authorize'
access_token_url = 'https://api.twitter.com/oauth/access_token'


def get_keys():
    registry = getUtility(IRegistry)
    try:
        key = registry['plone.twitter_consumer_key']
        secret = registry['plone.twitter_consumer_secret']
        return key, secret
    except:
        return None, None


def get_auth_url():
    key, secret = get_keys()
    callback_uri = api.portal.get().absolute_url() + '/@@twitter-auth'
    oauth_session = OAuth1Session(
        key, client_secret=secret, callback_uri=callback_uri)

    req = getRequest()
    token = oauth_session.fetch_request_token(request_token_url)
    req.response.setCookie('oauth_token_secret', token['oauth_token_secret'])

    return oauth_session.authorization_url(authorization_url)


def authorize():
    req = getRequest()
    key, secret = get_keys()
    callback_uri = api.portal.get().absolute_url() + '/@@twitter-callback'
    oauth_session = OAuth1Session(
        key, client_secret=secret, callback_uri=callback_uri)

    oauth_session.parse_authorization_response(
        req.URL + '?' + req.environ['QUERY_STRING'])
    return oauth_session.fetch_access_token(access_token_url)


def get_auth():
    registry = getUtility(IRegistry)
    check_keys = ('plone.twitter_consumer_key', 'plone.twitter_consumer_secret',
                  'plone.twitter_oauth_token', 'plone.twitter_oauth_secret')
    for key in check_keys:
        if not registry.get(key, None):
            return None
    try:
        return OAuth1(
            registry['plone.twitter_consumer_key'].encode('utf8'),
            client_secret=registry['plone.twitter_consumer_secret'].encode('utf8'),
            resource_owner_key=registry['plone.twitter_oauth_token'].encode('utf8'),
            resource_owner_secret=registry['plone.twitter_oauth_secret'].encode('utf8'),
            decoding=None)
    except:
        return None


STREAM_VERSION = '1.1'


class StreamListener(object):

    def __init__(self):
        pass

    def on_connect(self):
        """Called once connected to streaming server.

        This will be invoked once a successful response
        is received from the server. Allows the listener
        to perform some work prior to entering the read loop.
        """
        pass

    def on_data(self, raw_data):
        """Called when raw data is received from connection.

        Override this method if you wish to manually handle
        the stream data. Return False to stop stream and close connection.
        """
        pass

    def keep_alive(self):
        """Called when a keep-alive arrived"""
        return

    def on_exception(self, exception):
        """Called when an unhandled exception occurs."""
        return

    def on_limit(self, track):
        """Called when a limitation notice arrives"""
        return

    def on_error(self, status_code):
        """Called when a non-200 status code is returned"""
        return False

    def on_timeout(self):
        """Called when stream connection times out"""
        return

    def on_disconnect(self, notice):
        """Called when twitter sends a disconnect notice

        Disconnect codes are listed here:
        https://dev.twitter.com/docs/streaming-apis/messages#Disconnect_messages_disconnect
        """
        return

    def on_warning(self, notice):
        """Called when a disconnection warning message arrives"""
        return


class ReadBuffer(object):
    """Buffer data from the response in a smarter way than httplib/requests can.

    Tweets are roughly in the 2-12kb range, averaging around 3kb.
    Requests/urllib3/httplib/socket all use socket.read, which blocks
    until enough data is returned. On some systems (eg google appengine), socket
    reads are quite slow. To combat this latency we can read big chunks,
    but the blocking part means we won't get results until enough tweets
    have arrived. That may not be a big deal for high throughput systems.
    For low throughput systems we don't want to sacrafice latency, so we
    use small chunks so it can read the length and the tweet in 2 read calls.
    """

    def __init__(self, stream, chunk_size, encoding='utf-8'):
        self._stream = stream
        self._buffer = six.b('')
        self._chunk_size = chunk_size
        self._encoding = encoding

    def read_len(self, length):
        while not self._stream.closed:
            if len(self._buffer) >= length:
                return self._pop(length)
            read_len = max(self._chunk_size, length - len(self._buffer))
            self._buffer += self._stream.read(read_len)

    def read_line(self, sep=six.b('\n')):
        """Read the data stream until a given separator is found (default \n)

        :param sep: Separator to read until. Must by of the bytes type (str in python 2,
            bytes in python 3)
        :return: The str of the data read until sep
        """
        start = 0
        while not self._stream.closed:
            loc = self._buffer.find(sep, start)
            if loc >= 0:
                return self._pop(loc + len(sep))
            else:
                start = len(self._buffer)
            self._buffer += self._stream.read(self._chunk_size)

    def _pop(self, length):
        r = self._buffer[:length]
        self._buffer = self._buffer[length:]
        return r.decode(self._encoding)


class Stream(object):

    host = 'stream.twitter.com'

    def __init__(self, auth, listener=StreamListener(), **options):
        self.auth = auth
        self.listener = listener
        self.running = False
        self.timeout = options.get("timeout", 300.0)
        self.retry_count = options.get("retry_count")
        # values according to
        # https://dev.twitter.com/docs/streaming-apis/connecting#Reconnecting
        self.retry_time_start = options.get("retry_time", 5.0)
        self.retry_420_start = options.get("retry_420", 60.0)
        self.retry_time_cap = options.get("retry_time_cap", 320.0)
        self.snooze_time_step = options.get("snooze_time", 0.25)
        self.snooze_time_cap = options.get("snooze_time_cap", 16)

        # The default socket.read size. Default to less than half the size of
        # a tweet so that it reads tweets with the minimal latency of 2 reads
        # per tweet. Values higher than ~1kb will increase latency by waiting
        # for more data to arrive but may also increase throughput by doing
        # fewer socket read calls.
        self.chunk_size = options.get("chunk_size",  512)

        self.verify = options.get("verify", True)

        self.headers = options.get("headers") or {}
        self.new_session()
        self.body = None
        self.retry_time = self.retry_time_start
        self.snooze_time = self.snooze_time_step

    def new_session(self):
        self.session = requests.Session()
        self.session.headers = self.headers
        self.session.params = None

    def _run(self):
        # Authenticate
        url = "https://%s%s" % (self.host, self.url)

        # Connect and process the stream
        error_counter = 0
        resp = None
        exception = None
        while self.running:
            if self.retry_count is not None:
                if error_counter > self.retry_count:
                    # quit if error count greater than retry count
                    break
            try:
                resp = self.session.request('POST',
                                            url,
                                            data=self.body,
                                            timeout=self.timeout,
                                            stream=True,
                                            auth=self.auth,
                                            verify=self.verify)
                if resp.status_code != 200:
                    if self.listener.on_error(resp.status_code) is False:
                        break
                    error_counter += 1
                    if resp.status_code == 420:
                        self.retry_time = max(self.retry_420_start,
                                              self.retry_time)
                    sleep(self.retry_time)
                    self.retry_time = min(self.retry_time * 2,
                                          self.retry_time_cap)
                else:
                    error_counter = 0
                    self.retry_time = self.retry_time_start
                    self.snooze_time = self.snooze_time_step
                    self.listener.on_connect()
                    self._read_loop(resp)
            except (Timeout, ssl.SSLError) as exc:
                # This is still necessary, as a SSLError can actually be
                # thrown when using Requests
                # If it's not time out treat it like any other exception
                if isinstance(exc, ssl.SSLError):
                    if not (exc.args and 'timed out' in str(exc.args[0])):
                        exception = exc
                        break
                if self.listener.on_timeout() is False:
                    break
                if self.running is False:
                    break
                sleep(self.snooze_time)
                self.snooze_time = min(self.snooze_time + self.snooze_time_step,
                                       self.snooze_time_cap)
            except Exception as exc:
                exception = exc
                # any other exception is fatal, so kill loop
                break

        # cleanup
        self.running = False
        if resp:
            resp.close()

        self.new_session()

        if exception:
            # call a handler first so that the exception can be logged.
            self.listener.on_exception(exception)
            raise exception

    def _data(self, data):
        if self.listener.on_data(data) is False:
            self.running = False

    def _read_loop(self, resp):
        charset = resp.headers.get('content-type', default='')
        enc_search = re.search('charset=(?P<enc>\S*)', charset)
        if enc_search is not None:
            encoding = enc_search.group('enc')
        else:
            encoding = 'utf-8'

        buf = ReadBuffer(resp.raw, self.chunk_size, encoding=encoding)

        while self.running and not resp.raw.closed:
            length = 0
            while not resp.raw.closed:
                line = buf.read_line().strip()
                if not line:
                    self.listener.keep_alive()  # keep-alive new lines are expected
                elif line.isdigit():
                    length = int(line)
                    break
                else:
                    raise Exception('Expecting length, unexpected value found')

            next_status_obj = buf.read_len(length)
            if self.running:
                self._data(next_status_obj)

            # # Note: keep-alive newlines might be inserted before each length value.
            # # read until we get a digit...
            # c = b'\n'
            # for c in resp.iter_content(decode_unicode=True):
            #     if c == b'\n':
            #         continue
            #     break
            #
            # delimited_string = c
            #
            # # read rest of delimiter length..
            # d = b''
            # for d in resp.iter_content(decode_unicode=True):
            #     if d != b'\n':
            #         delimited_string += d
            #         continue
            #     break
            #
            # # read the next twitter status object
            # if delimited_string.decode('utf-8').strip().isdigit():
            #     status_id = int(delimited_string)
            #     next_status_obj = resp.raw.read(status_id)
            #     if self.running:
            #         self._data(next_status_obj.decode('utf-8'))

        if resp.raw.closed:
            self.on_closed(resp)

    def _start(self, async):
        self.running = True
        if async:
            self._thread = Thread(target=self._run)
            self._thread.start()
        else:
            self._run()

    def on_closed(self, resp):
        """ Called when the response has been closed by Twitter """
        pass

    def userstream(self,
                   stall_warnings=False,
                   _with=None,
                   replies=None,
                   track=None,
                   locations=None,
                   async=False,
                   encoding='utf8'):
        self.session.params = {'delimited': 'length'}
        if self.running:
            raise Exception('Stream object already connected!')
        self.url = '/%s/user.json' % STREAM_VERSION
        self.host = 'userstream.twitter.com'
        if stall_warnings:
            self.session.params['stall_warnings'] = stall_warnings
        if _with:
            self.session.params['with'] = _with
        if replies:
            self.session.params['replies'] = replies
        if locations and len(locations) > 0:
            if len(locations) % 4 != 0:
                raise Exception("Wrong number of locations points, "
                                "it has to be a multiple of 4")
            self.session.params['locations'] = ','.join(['%.2f' % l for l in locations])
        if track:
            self.session.params['track'] = u','.join(track).encode(encoding)

        self._start(async)

    def firehose(self, count=None, async=False):
        self.session.params = {'delimited': 'length'}
        if self.running:
            raise Exception('Stream object already connected!')
        self.url = '/%s/statuses/firehose.json' % STREAM_VERSION
        if count:
            self.url += '&count=%s' % count
        self._start(async)

    def retweet(self, async=False):
        self.session.params = {'delimited': 'length'}
        if self.running:
            raise Exception('Stream object already connected!')
        self.url = '/%s/statuses/retweet.json' % STREAM_VERSION
        self._start(async)

    def sample(self, async=False, languages=None):
        self.session.params = {'delimited': 'length'}
        if self.running:
            raise Exception('Stream object already connected!')
        self.url = '/%s/statuses/sample.json' % STREAM_VERSION
        if languages:
            self.session.params['language'] = ','.join(map(str, languages))
        self._start(async)

    def filter(self, follow=None, track=None, async=False, locations=None,
               stall_warnings=False, languages=None, encoding='utf8', filter_level=None):
        self.body = {}
        self.session.headers['Content-type'] = "application/x-www-form-urlencoded"
        if self.running:
            raise Exception('Stream object already connected!')
        self.url = '/%s/statuses/filter.json' % STREAM_VERSION
        if follow:
            self.body['follow'] = u','.join(follow).encode(encoding)
        if track:
            self.body['track'] = u','.join(track).encode(encoding)
        if locations and len(locations) > 0:
            if len(locations) % 4 != 0:
                raise Exception("Wrong number of locations points, "
                                "it has to be a multiple of 4")
            self.body['locations'] = u','.join(['%.4f' % l for l in locations])
        if stall_warnings:
            self.body['stall_warnings'] = stall_warnings
        if languages:
            self.body['language'] = u','.join(map(str, languages))
        if filter_level:
            self.body['filter_level'] = unicode(filter_level, encoding)
        self.session.params = {'delimited': 'length'}
        self.host = 'stream.twitter.com'
        self._start(async)

    def sitestream(self, follow, stall_warnings=False,
                   with_='user', replies=False, async=False):
        self.body = {}
        if self.running:
            raise Exception('Stream object already connected!')
        self.url = '/%s/site.json' % STREAM_VERSION
        self.body['follow'] = u','.join(map(six.text_type, follow))
        self.body['delimited'] = 'length'
        if stall_warnings:
            self.body['stall_warnings'] = stall_warnings
        if with_:
            self.body['with'] = with_
        if replies:
            self.body['replies'] = replies
        self._start(async)

    def disconnect(self):
        if self.running is False:
            return
        self.running = False



from castle.cms import authentication
from castle.cms import cache
from castle.cms import texting
from castle.cms.interfaces import IAuthenticator
from castle.cms.interfaces import ISecureLoginAllowedView
from castle.cms.interfaces import ISiteSchema
from castle.cms.utils import get_managers
from castle.cms.utils import send_email
from castle.cms.utils import strings_differ
from plone import api
from plone.protect.authenticator import createToken
from plone.protect.interfaces import IDisableCSRFProtection
from plone.registry.interfaces import IRegistry
from Products.Five import BrowserView
from Products.PasswordResetTool.PasswordResetTool import ExpiredRequestError
from Products.PasswordResetTool.PasswordResetTool import InvalidRequestError
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.component.interfaces import ComponentLookupError
from zope.interface import alsoProvides
from zope.interface import implements

import json
import time


class SecureLoginView(BrowserView):
    implements(ISecureLoginAllowedView)

    def __init__(self, context, request):
        super(SecureLoginView, self).__init__(context, request)
        self.auth = self.authenticator = getMultiAdapter(
            (context, request), IAuthenticator)

    def __call__(self):
        api_method = self.request.form.get('apiMethod')
        if api_method:
            self.request.response.setHeader('Content-type', 'application/json')

        if api_method == 'send_authorization':
            return self.send_authorization()
        if api_method == 'authorize_code':
            return self.authorize_code()
        if api_method == 'login':
            return self.login()
        if api_method == 'set_password':
            return self.set_password()
        if api_method == 'request_country_exception':
            return self.request_country_exception()
        if api_method == 'reset_password':
            return self.reset_password()

        self.request.response.setHeader('X-Theme-Applied', True)
        return self.index()

    @property
    def username(self):
        return self.request.form.get('username')

    def get_country_header(self):
        return (
            self.request.get_header('Cf-Ipcountry') or
            self.request.get_header('Country'))

    def request_country_exception(self):
        # re-check login...
        # verify we get to country block again
        try:
            self.auth.authenticate(
                username=self.username,
                password=self.request.form.get('password'),
                country=self.get_country_header())
            return json.dumps({
                'success': False,
                'message': 'Error authenticating request',
                'messageType': 'error'
            })
        except authentication.AuthenticationCountryBlocked:
            pass

        req_country = self.get_country_header()
        user = api.user.get(username=self.username)
        data = self.auth.issue_country_exception(user, req_country)

        email_subject = "Country block exception request(Site: %s)" % (
            api.portal.get_registry_record('plone.site_title'))

        for admin_user in get_managers():
            if admin_user.getId() == user.getId():
                # it could be an admin, do not allow him to authorize himself
                continue

            admin_email = admin_user.getProperty('email')

            email_data = data.copy()
            email_data.update({
                'admin_name': (admin_user.getProperty('fullname') or
                               admin_user.getUserName()),
                'name': (user.getProperty('fullname') or user.getUserName()),
                'auth_link': '{}/authorize-country-login?code={}&userid={}'.format(
                    self.context.absolute_url(),
                    data['code'],
                    user.getId()
                )
            })
            email_html = '''
<p>
Hello {admin_name},
</p>
<p>
There has been a request to suspend country login restrictions for this site.

<p>
The user requesting this access logged this information:
</p>
<ul>
<li><b>Username</b>: {username}</li>
<li><b>Name</b>: {name}</li>
<li><b>Country</b>: {country}</li>
<li><b>IP</b>: {ip}</li>
<li><b>User agent</b>: {user_agent}</li>
</ul>

<p>The user has 12 hours between after authorization has been giving to login
   with the temporary access</p>

<p>If you'd like to authorize this user, please click this link:
   <a href="{auth_link}">authorize {username}</a>
</p>'''.format(**email_data)
            send_email(admin_email, email_subject, html=email_html)

        return json.dumps({
            'success': True,
            'message': 'Successfully requested country exception.',
            'messageType': 'info'
        })

    def set_password(self):
        # 1. only set password for logged in user...
        if api.user.is_anonymous():
            return json.dumps({
                'success': False,
                'message': 'Need to be logged in to reset password'
            })

        acl_users = self.get_tool('acl_users')

        # 2. double check existing password
        user = acl_users.authenticate(
            self.username, self.request.form.get('existing_password'),
            self.request)

        if not user:
            return json.dumps({
                'success': False,
                'message': 'Existing password did not match'
            })

        if user.getId() != api.user.get_current().getId():
            return json.dumps({
                'success': False,
                'message': 'Error resetting'
            })

        # 3. finally, set password
        newpw = self.request.form.get('new_password')

        mtool = api.portal.get_tool('portal_membership')
        member = mtool.getMemberById(user.getId())

        self.auth.change_password(member, newpw)

        return json.dumps({
            'success': True
        })

    def login(self):
        # double check auth code first
        code = self.request.form.get('code')
        if (self.auth.two_factor_enabled and not
                self.auth.authorize_2factor(self.username, code, 5 * 60)):
            return json.dumps({
                'success': False,
                'message': 'Login failed for username and password.'
            })

        try:
            authorized, user = self.auth.authenticate(
                username=self.username,
                password=self.request.form.get('password'),
                country=self.get_country_header())
            if authorized:
                reset_password = user.getProperty(
                    'reset_password_required', False)

                resp = {
                    'success': True,
                    'resetpassword': reset_password
                }
                try:
                    with api.env.adopt_user(user=user):
                        resp['authenticator'] = createToken()
                        return json.dumps(resp)
                except:
                    resp['authenticator'] = createToken()
                    return json.dumps(resp)
            else:
                return json.dumps({
                    'success': False,
                    'message': 'Login failed for username and password.'
                })
        except authentication.AuthenticationMaxedLoginAttempts:
            return json.dumps({
                'success': False,
                'message': 'You have reached the have number of login attempts '
                           'for a period.'
            })
        except authentication.AuthenticationUserDisabled:
            return json.dumps({
                'success': False,
                'message': 'User is disabled.'
            })
        except authentication.AuthenticationCountryBlocked:
            return json.dumps({
                'success': True,
                'countryBlocked': True,
                'message': 'User login blocked. The country you are logging '
                           'in from is blocked.'
            })
        except authentication.AuthenticationPasswordResetWindowExpired:
            return json.dumps({
                'success': False,
                'message': 'User login disabled. Password reset request not '
                           'fullfilled in required time period.'
            })

    def authorize_code(self):
        code = self.request.form.get('code')
        if self.authenticator.authorize_2factor(self.username, code):
            return json.dumps({
                'success': True,
                'message': 'Authorization code verified.'
            })
        else:
            return json.dumps({
                'success': False,
                'message': 'Failed to authorize code'
            })

    def send_authorization(self):
        auth_type = self.request.form.get('authType') or 'email'
        if auth_type == 'email':
            self.send_auth_email()
        elif auth_type == 'sms':
            self.send_auth_text()
        return json.dumps({
            'success': True,
            'message': 'Authorization code sent to provided username if '
                       'username exists.'
        })

    def send_auth_email(self):
        email = None
        user = api.user.get(username=self.username)
        if user:
            email = user.getProperty('email')
        if not email:
            return

        code = self.auth.issue_2factor_code(self.username)
        registry = getUtility(IRegistry)
        site_settings = registry.forInterface(ISiteSchema,
                                              prefix="plone",
                                              check=False)
        html = """
<p>
    You have requested to authorize access to {title}.
</p>
<p>
    You authorization code is: {code}
</p>""".format(title=site_settings.site_title,
               code=code)

        send_email(
            [email], "Authorization code(%s)" % site_settings.site_title,
            html=html)

    def send_auth_text(self):
        phone = None
        user = api.user.get(username=self.username)
        if user:
            phone = user.getProperty('phone_number')

        if not phone:
            return

        registry = getUtility(IRegistry)
        site_settings = registry.forInterface(ISiteSchema,
                                              prefix="plone",
                                              check=False)

        code = self.auth.issue_2factor_code(self.username)

        text_message = '{}: phone verification code: {}'.format(
            site_settings.site_title, code)

        return texting.send(text_message, phone)

    def reset_password(self):
        pw_tool = api.portal.get_tool('portal_password_reset')
        userid = self.request.form.get('userid')
        randomstring = self.request.form.get('code')
        password = self.request.form.get('password')
        alsoProvides(self.request, IDisableCSRFProtection)
        try:
            pw_tool.resetPassword(userid, randomstring, password)
        except ExpiredRequestError:
            return json.dumps({
                'success': False,
                'message': 'Password reset request has expired'
            })
        except (InvalidRequestError, RuntimeError):
            return json.dumps({
                'success': False,
                'message': 'Password reset request is invalid'
            })

        return json.dumps({
            'success': True,
            'message': 'Password reset successfully'
        })

    def get_registry(self):
        try:
            return getUtility(IRegistry)
        except ComponentLookupError:
            pass

    def get_tool(self, name):
        if self.auth.is_zope_root:
            return getattr(self.context, name, None)
        else:
            return api.portal.get_tool(name)

    def options(self):
        site_url = self.context.absolute_url()
        data = {
            'supportedAuthSchemes': self.authenticator.get_supported_auth_schemes(),
            'twoFactorEnabled': self.authenticator.two_factor_enabled,
            'apiEndpoint': '{}/@@secure-login'.format(site_url),
            'successUrl': site_url,
            'authenticator': createToken()
        }

        username = None
        pwreset = self.request.form.get('pwreset') == 'true'
        if pwreset:
            try:
                user = api.user.get(self.request.form.get('userid'))
                username = user.getUserName()
                data.update({
                    'passwordReset': pwreset,
                    'username': username,
                    'code': self.request.form.get('code'),
                    'userid': self.request.form.get('userid')
                })
            except:
                pwreset = False
        return json.dumps(data)


class LoginExceptionApprovalView(BrowserView):
    implements(ISecureLoginAllowedView)

    message = 'Incorrect code for country exception.'
    success = False

    def __call__(self):
        auth = self.authenticator = getMultiAdapter(
            (self.context, self.request), IAuthenticator)

        userid = self.request.form.get('userid')
        code = self.request.form.get('code')
        if userid and code:
            exc_key = auth.get_country_exception_cache_key(userid)
            try:
                data = cache.get(exc_key)
                if not strings_differ(data['code'], code):
                    timestamp = data.get('timestamp')
                    if timestamp and (time.time() < (timestamp + (12 * 60 * 60))):
                        user = api.user.get(data['userid'])
                        self.message = 'Successfully issued country login exception for {}({}).'.format(  # noqa
                            user.getProperty('fullname') or user.getUserName(),
                            user.getUserName())
                        self.success = True
                        data['granted'] = True
                        data['timestamp'] = time.time()
                        cache.set(exc_key, data, 12 * 60 * 60)
                        self.send_email(data)
            except:
                pass
        return self.index()

    def send_email(self, user, data):
        email_subject = "Country block exception request granted(Site: %s)" % (
            api.portal.get_registry_record('plone.site_title'))

        email = user.getProperty('email')

        email_data = data.copy()
        email_data.update({
            'name': (user.getProperty('fullname') or user.getUserName())
        })
        email_html = '''
<p>
Hello {name},
</p>
<p>
Your request to login from your current location is granted.
<br />
Please login again with the same browser and location you made the request.
If you still experience any problems, please contact your administrator.

<p>
User and location information:
</p>
<ul>
<li><b>Username</b>: {username}</li>
<li><b>Name</b>: {name}</li>
<li><b>Country</b>: {country}</li>
<li><b>IP</b>: {ip}</li>
<li><b>User agent</b>: {user_agent}</li>
</ul>

<p>You have 12 hours to use this granted login exception.</p>
'''.format(**email_data)
        send_email(email, email_subject, html=email_html)



from castle.cms.constants import CRAWLED_DATA_KEY
from castle.cms.constants import CRAWLED_SITE_ES_DOC_TYPE
from castle.cms.interfaces import ICrawlerConfiguration
from collective.elasticsearch.es import ElasticSearchCatalog
from elasticsearch import TransportError
from plone import api
from plone.app.registry.browser import controlpanel
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.annotation.interfaces import IAnnotations


class CrawlerControlPanelForm(controlpanel.RegistryEditForm):
    schema_prefix = 'castle'
    schema = ICrawlerConfiguration
    id = "CrawlerControlPanel"
    label = u"Crawler Configuration"


class CrawlerControlPanel(controlpanel.ControlPanelFormWrapper):
    form = CrawlerControlPanelForm
    index = ViewPageTemplateFile('templates/crawler.pt')

    def get_crawl_data(self):
        annotations = IAnnotations(self.context)
        if CRAWLED_DATA_KEY not in annotations:
            return {
                'tracking': {}
            }
        return annotations[CRAWLED_DATA_KEY]

    def get_index_summary(self):
        query = {
            "size": 0,
            "aggregations": {
                "totals": {
                    "terms": {
                        "field": "domain"
                    }
                }
            }
        }
        portal_catalog = api.portal.get_tool('portal_catalog')
        try:
            es = ElasticSearchCatalog(portal_catalog)
            result = es.connection.search(
                index=es.index_name,
                doc_type=CRAWLED_SITE_ES_DOC_TYPE,
                body=query)
        except TransportError:
            return []

        data = result['aggregations']['totals']['buckets']
        return data



from castle.cms import trash
from castle.cms.interfaces import ITrashed
from plone import api
from plone.app.linkintegrity.exceptions import LinkIntegrityNotificationException
from plone.locking.interfaces import ILockable
from Products.Five import BrowserView
from unidecode import unidecode


class TrashView(BrowserView):

    def __call__(self):
        self.catalog = api.portal.get_tool('portal_catalog')
        self.site = api.portal.get()
        self.site_path = '/'.join(self.site.getPhysicalPath())
        if self.request.REQUEST_METHOD == 'POST':
            action = self.request.get('action')
            if action == 'restore':
                self.restore()
            elif action == 'delete':
                self.delete()
            elif action == 'empty':
                self.empty()
        self.items = self.catalog(trashed=True, sort_on='modified',
                                  object_provides=ITrashed.__identifier__)
        return self.index()

    def get_label(self, brain):
        label = brain.Title
        if brain.is_folderish:
            num = len(self.catalog(trashed=True,
                                   path={'query': brain.getPath(), 'depth': -1}))
            label += '(%i items)' % num
        return label

    def get_path(self, brain):
        if hasattr(brain, 'getPath'):
            path = brain.getPath()
        else:
            path = '/'.join(brain.getPhysicalPath())
        return path[len(self.site_path):]

    def get_by_uid(self, uid):
        return self.catalog(UID=uid, trashed=True)[0].getObject()

    def restore(self):
        uid = self.request.get('uid')
        obj = self.get_by_uid(uid)
        api.portal.show_message(u'Successfully restored "%s" located at: %s' % (
            unidecode(obj.Title()), self.get_path(obj)), self.request, type='info')
        trash.restore(obj)

    def delete(self):
        obj = self.get_by_uid(self.request.get('uid'))
        lockable = ILockable(obj, None)
        if lockable and lockable.locked():
            lockable.clear_locks()
        try:
            api.content.delete(obj)
            api.portal.show_message('Successfully deleted "%s" located at: %s' % (
                unidecode(obj.Title()), self.get_path(obj)), self.request, type='warning')
        except LinkIntegrityNotificationException:
            api.portal.show_message('Can not delete "%s" located at: %s because it is still linked.' % (
                unidecode(obj.Title()), self.get_path(obj)), self.request, type='warning')

    def empty(self):
        for item in [i for i in self.catalog(trashed=True,
                                             object_provides=ITrashed.__identifier__)]:
            obj = item.getObject()
            lockable = ILockable(obj, None)
            if lockable and lockable.locked():
                lockable.clear_locks()
            try:
                api.content.delete(obj, check_linkintegrity=False)
            except LinkIntegrityNotificationException:
                # could be a folder that has been deleted
                api.portal.show_message(
                    'Some content could not be removed because it is still linked '
                    'to other content on the site.',
                    self.request, type='warning')
        api.portal.show_message('Trash emptied', self.request, type='warning')



import logging

from Acquisition import aq_inner
from Acquisition import aq_parent
from Products.CMFPlone.interfaces import IPloneSiteRoot
from castle.cms.interfaces import IGlobalTile
from castle.cms.interfaces import IMetaTile
from plone.supermodel import model
from plone.tiles import Tile
from zope import schema
from zope.component import getMultiAdapter
from zope.interface import alsoProvides
from zope.interface import implements
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


logger = logging.getLogger('castle.cms')


class MetaTile(Tile):
    implements(IGlobalTile)

    _template = """<section class="meta-tile-container"
                        id="meta-tile-%(id)s">%(content)s</section>"""
    _tile_template = """<div class="tile-container
                                    tile-container-%(type)s
                                    tile-container-index-%(idx)i"
                             id="tile-%(id)s">%(content)s</div>"""

    def render_meta_tile(self, meta_tile, idx_start=0):
        if not meta_tile:
            return ''
        if meta_tile.data.get('mode') == 'hide':
            return ''
        tiles = meta_tile.data.get('tiles') or []
        rendered = ''
        for idx, tile_config in enumerate(tiles):
            tile = getMultiAdapter((self.context, self.request),
                                   name=tile_config['type'])
            alsoProvides(tile, IMetaTile)
            tile.id = tile_config['id']
            tile.__name__ = tile_config['type']
            tile.data_context = meta_tile.context
            tile.wrap = False
            result = tile()
            if result:
                result = result.replace(
                    '<html><body>', '').replace('</body></html>', '').strip()
                if result:
                    rendered += self._tile_template % {
                        'id': tile.id,
                        'content': result,
                        'type': tile.__name__.replace('.', '-'),
                        'idx': idx
                    }
        return rendered

    def index(self):
        meta_tiles = getEffectiveTiles(self.context, self.request, self.id)
        result = ''
        for meta_tile in meta_tiles:
            result += self.render_meta_tile(meta_tile)

        if result:
            result = self._template % {
                'id': self.id,
                'content': result
            }
        return result

    def __call__(self):
        try:
            return self.index()
        except:
            logger.warn('Error rendering meta tile', exc_info=True)
            return ''


class IMetaTileSchema(model.Schema):

    # dict structure:
    #   id: < generated id of file >
    #   type type: < type of tile >
    tiles = schema.List(
        title=u'Tiles',
        value_type=schema.Dict(value_type=schema.Field(),
                               key_type=schema.TextLine()),
        required=False,
        default=[],
        missing_value=[]
    )

    mode = schema.Choice(
        title=u"Mode",
        vocabulary=SimpleVocabulary([
            SimpleTerm('hide', 'hide', u'Hide'),
            SimpleTerm('parent', 'parent', u'Use parent settings'),
            SimpleTerm('show', 'show', u'Show(hides parent tiles)'),
            SimpleTerm('add', 'add', u'Add')
        ]),
        default='parent'
    )


def getEffectiveTiles(ob, req, _id, parent_id=None):
    if parent_id is None:
        parent_id = _id
    tile = getTile(ob, req, _id)
    try:
        data = tile.data
    except TypeError:
        return []
    tiles = []
    mode = data.get('mode') or 'parent'
    while not IPloneSiteRoot.providedBy(ob):
        if mode != 'parent' and mode != 'add':
            break
        if mode == 'add':
            tiles.append(tile)
        ob = aq_parent(ob)
        tile = getTile(ob, req, parent_id)
        mode = tile.data.get('mode') or 'parent'
    tiles.append(tile)
    return list(reversed(tiles))


def getTile(ob, req, _id):
    tile = MetaTile(aq_inner(ob), req)
    alsoProvides(tile, IGlobalTile)
    tile.id = _id
    tile.__name__ = 'castle.cms.meta'
    return tile



from plone.autoform.interfaces import IFormFieldProvider
from plone.dexterity.interfaces import IDexterityContent
from plone.supermodel import model
from zope import schema
from zope.component import adapter
from zope.interface import alsoProvides
from zope.interface import implementer


class INavigationLabel(model.Schema):

    model.fieldset(
        'settings',
        fields=['navigation_label'],
    )

    navigation_label = schema.TextLine(
        title=u'Navigation Label',
        description=u'Label that shows up in Navigation menus instead of title. '
                    u'Can be left blank',
        required=False
    )

alsoProvides(INavigationLabel, IFormFieldProvider)


@implementer(INavigationLabel)
@adapter(IDexterityContent)
class NavigationLabel(object):

    def __init__(self, context):
        self.context = context



from castle.cms.tiles.base import BaseTile
from plone import api
from plone.registry.interfaces import IRegistry
from plone.tiles.interfaces import IPersistentTile
from zope import schema
from zope.component import getUtility
from zope.interface import implements
from zope.interface import Interface


class EmbedTile(BaseTile):
    implements(IPersistentTile)

    def render(self):
        content = self.data.get('code', '') or ''
        site_url = portal_url = api.portal.get_tool('portal_url')
        site_url = portal_url()
        registry = getUtility(IRegistry)
        public_url = registry.get('plone.public_url', None)
        if not public_url:
            public_url = site_url
            if not api.user.is_anonymous():
                site_url = public_url
        context_url = self.context.absolute_url()
        return content.replace(
            '{context-url}', context_url).replace(
            '{public-url}', public_url).replace(
            '{site-url}', site_url)


class IEmbedTileSchema(Interface):

    code = schema.Text(
        title=u"Code",
        description=u"Raw html code to include"
    )



from castle.cms.interfaces import IDashboard
from castle.cms.utils import is_mosaic_edit_form
from plone import api
from plone.app.blocks.browser import ContentLayoutView
from plone.app.layout.globals.layout import LayoutPolicy as BaseLayoutPolicy
from plone.i18n.normalizer.interfaces import IIDNormalizer
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces import ILinkSchema
from Products.Five.browser.metaconfigure import ViewMixinForTemplates
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile as ZopeViewPageTemplateFile
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.component import queryUtility


class LayoutView(ContentLayoutView):

    def __call__(self):
        if IDashboard.providedBy(self.context) and '@@dashboard' not in self.request.URL:
            return self.request.response.redirect(
                api.portal.get().absolute_url() + '/@@dashboard'
            )
        return super(LayoutView, self).__call__()


class LayoutPolicy(BaseLayoutPolicy):

    def have_portlets(self, manager_name, view=None):
        return False

    def bodyClass(self, template, view):
        """
        Returns the CSS class to be used on the body tag.

        Included body classes
        - template name: template-{}
        - portal type: portaltype-{}
        - navigation root: site-{}
        - section: section-{}
            - only the first section
        - section structure
            - a class for every container in the tree
        - hide icons: icons-on
        - markspeciallinks: pat-markspeciallinks
        - userrole-{} for each role the user has in this context
        - min-view-role: role required to view context
        - default-view: if view is the default view
        """
        context = self.context
        portal_state = getMultiAdapter(
            (context, self.request), name=u'plone_portal_state')
        normalizer = queryUtility(IIDNormalizer)

        body_classes = []
        # template class (required)
        name = ''
        if isinstance(template, ViewPageTemplateFile) or \
           isinstance(template, ZopeViewPageTemplateFile) or \
           isinstance(template, ViewMixinForTemplates):
            # Browser view
            name = view.__name__
        elif template is not None:
            name = template.getId()
        if name:
            name = normalizer.normalize(name)
            body_classes.append('template-%s' % name)

        # portal type class (optional)
        portal_type = normalizer.normalize(context.portal_type)
        if portal_type:
            body_classes.append("portaltype-%s" % portal_type)

        # section class (optional)
        navroot = portal_state.navigation_root()
        body_classes.append("site-%s" % navroot.getId())

        contentPath = context.getPhysicalPath()[
            len(navroot.getPhysicalPath()):]
        if contentPath:
            body_classes.append("section-%s" % contentPath[0])
            # skip first section since we already have that...
            if len(contentPath) > 1:
                registry = getUtility(IRegistry)
                depth = registry.get(
                    'plone.app.layout.globals.bodyClass.depth', 4)
                if depth > 1:
                    classes = ['subsection-%s' % contentPath[1]]
                    for section in contentPath[2:depth]:
                        classes.append('-'.join([classes[-1], section]))
                    body_classes.extend(classes)

        # class for hiding icons (optional)
        if self.icons_visible():
            body_classes.append('icons-on')
        else:
            body_classes.append('icons-off')

        # class for hiding thumbs (optional)
        if self.thumb_visible():
            body_classes.append('thumbs-on')
        else:
            body_classes.append('thumbs-off')

        # permissions required. Useful to theme frontend and backend
        # differently
        permissions = []
        if not getattr(view, '__ac_permissions__', tuple()):
            permissions = ['none']
        for permission, roles in getattr(view, '__ac_permissions__', tuple()):
            permissions.append(normalizer.normalize(permission))
        if 'none' in permissions or 'view' in permissions:
            body_classes.append('frontend')
        for permission in permissions:
            body_classes.append('viewpermission-' + permission)

        # class for user roles
        membership = getToolByName(context, "portal_membership")

        if membership.isAnonymousUser():
            body_classes.append('userrole-anonymous')
        else:
            user = membership.getAuthenticatedMember()
            context_roles = user.getRolesInContext(self.context)
            for role in context_roles:
                body_classes.append('userrole-' + role.lower().replace(' ', '-'))

            if len(set(context_roles) - set(['Authenticated', 'Member'])) == 0:
                body_classes.append('no-roles')

            # toolbar classes
            if not is_mosaic_edit_form(self.request):
                body_classes.append('castle-toolbar-active')

        # class for markspeciallinks pattern
        registry = getUtility(IRegistry)
        settings = registry.forInterface(ILinkSchema,
                                         prefix="plone",
                                         check=False)

        msl = settings.mark_special_links
        elonw = settings.external_links_open_new_window
        if msl or elonw:
            body_classes.append('pat-markspeciallinks')

        return ' '.join(body_classes)



# check failed login attempts
# lockout user if more than 5 for 15 minutes, send system message
from castle.cms import cache
from castle.cms.utils import get_ip
from datetime import datetime
from plone import api
from plone.memoize.instance import memoize
from plone.registry.interfaces import IRegistry
from time import time
from uuid import uuid4
from zope.component import getUtility

import cPickle
import logging


LOGGED_IN_MARKER_KEY = 'castle.cms.user-logged-in'

log = logging.getLogger(__name__)


class LockoutManager(object):

    def __init__(self, site, login):
        self.site = site
        self.login = login

        try:
            self.attempts = cache.get(self.cache_key)
        except (AttributeError, KeyError):
            self.attempts = []

        registry = getUtility(IRegistry)
        self.window = registry.get('plone.failed_login_attempt_window', 3600)
        self.max_number = registry.get('plone.max_failed_login_attempts', 15)

    @property
    @memoize
    def cache_key(self):
        return '%s-loginattempts-%s' % (
            '-'.join(self.site.getPhysicalPath()[1:]),
            self.login
        )

    def get_attempts_this_window(self):
        now = time()
        startperiod = now - self.window
        period_attempts = []
        for tt in self.attempts:
            if tt > startperiod and tt <= now:
                period_attempts.append(tt)
        return period_attempts

    def add_attempt(self):
        attempts = self.get_attempts_this_window()
        attempts.append(time())
        self.attempts = attempts
        self.save()

    def maxed_number_of_attempts(self):
        attempts = self.get_attempts_this_window()
        return len(attempts) >= self.max_number

    def clear(self):
        self.attempts = []
        self.save()

    def save(self):
        cache.set(self.cache_key, self.attempts)


class SessionManager(object):
    cookie_name = 'castle_session_id'

    def __init__(self, site, request, user):
        self.user = user
        self.site = site
        self.request = request
        self.session_id = self.request.cookies.get(self.cookie_name, None)

    def has_session_id(self):
        return self.session_id is not None

    @property
    @memoize
    def cache_key(self):
        return '%s-session-%s-%s' % (
            '-'.join(self.site.getPhysicalPath()[1:]),
            self.user.getId(),
            self.session_id
        )

    def register(self):
        self.session_id = uuid4()
        self.request.response.setCookie(self.cookie_name, self.session_id)
        self.log({})

    def get(self, cache_key=None):
        try:
            return cache.get(self.cache_key)
        except (AttributeError, KeyError):
            return None

    def log(self, session):
        session.update({
            'updated': datetime.utcnow().isoformat(),
            'ua': self.request.environ.get('HTTP_USER_AGENT', 'unknown'),
            'ip': get_ip(self.request),
            'userid': self.user.getId(),
            'id': self.session_id
        })
        # default of 5 hr sessions
        # ideally this matches cookie timeout
        cache.set(self.cache_key, session, expire=60 * 60 * 5)

    def expired(self, session):
        try:
            return session.get('expired', False)
        except:
            return False

    def delete(self):
        try:
            cache.delete(self.cache_key)
        except:
            pass

    def expire(self, session=None):
        if session is None:
            session = self.get()

        if session:
            session.update({
                'expired': True
            })
            cache.set(self.cache_key, session, expire=60 * 60 * 5)


def get_active_sessions(sessions_key=None):
    portal = api.portal.get()
    portal_path = '-'.join(portal.getPhysicalPath()[1:])
    if not sessions_key:
        sessions_key = '%s-session-*' % portal_path

    cclient = cache.get_client()
    sessions = []
    try:
        keys = cclient.client.keys(sessions_key)
        if keys:
            for session in cclient.client.mget(keys):
                try:
                    session = cPickle.loads(session)
                except:
                    continue
                sessions.append(session)
            sessions = reversed(sorted(sessions, key=lambda x: x['updated']))
    except AttributeError:
        sessions = []

    # we'll also squash and cleanup duplicate sessions here...
    found = []
    filtered = []
    for session in sessions:
        compare_key = '%s%s%s' % (
            session.get('ua'),
            session.get('ip'),
            session.get('userid')
        )
        if compare_key in found:
            cache_key = '%s-session-%s-%s' % (
                portal_path,
                session['userid'],
                session['id']
            )
            try:
                cache.delete(cache_key)
            except:
                pass
        else:
            try:
                session['user'] = api.user.get(session['userid'])
            except:
                pass
            found.append(compare_key)
            filtered.append(session)
    return filtered



from castle.cms.behaviors.search import ISearch
from castle.cms.social import COUNT_ANNOTATION_KEY
from collective.elasticsearch import mapping
from collective.elasticsearch import query
from collective.elasticsearch.interfaces import IAdditionalIndexDataProvider
from plone import api
from zope.annotation.interfaces import IAnnotations
from zope.interface import implements


class MappingAdapter(mapping.MappingAdapter):
    _default_mapping = mapping.MappingAdapter._default_mapping.copy()
    _default_mapping.update({
        'page_views': {'store': True, 'type': 'integer', 'null_value': 0},
        'facebook_shares': {'store': True, 'type': 'integer', 'null_value': 0},
        'twitter_shares': {'store': True, 'type': 'integer', 'null_value': 0},
        'linkedin_shares': {'store': True, 'type': 'integer', 'null_value': 0},
        'pinterist_shares': {'store': True, 'type': 'integer', 'null_value': 0},
        'searchterm_pins': {'store': True, 'type': 'string',
                            'index': 'not_analyzed', 'null_value': '[]'},
        'contributors': {'store': False, 'type': 'string',
                         'index': 'not_analyzed', 'null_value': '[]'},
        'immediate_folder': {'store': True, 'type': 'string',
                             'index': 'not_analyzed', 'null_value': ''},
        'parent_folder': {'store': True, 'type': 'string',
                          'index': 'not_analyzed', 'null_value': ''}
    })


class AdditionalIndexDataProvider(object):
    implements(IAdditionalIndexDataProvider)

    def __init__(self, obj):
        self.obj = obj

    def __call__(self, es, existing_data):
        annotations = IAnnotations(self.obj)
        data = {}
        counts = annotations.get(COUNT_ANNOTATION_KEY, {})
        for key, value in counts.items():
            if key in ('page_views',):
                data[key] = value
            else:
                data[key + '_shares'] = value
        if 'twitter' in counts:
            tdata = counts['twitter']
            data['twitter_shares'] = tdata.get('total', 0)
        sdata = ISearch(self.obj, None)
        if sdata:
            data['searchterm_pins'] = [t.lower() for t in sdata.searchterm_pins or [] if t]
        else:
            data['searchterm_pins'] = []

        try:
            data['SearchableText'] = u'%s %s' % (
                existing_data.get('SearchableText', ''),
                u' '.join(data['searchterm_pins']))
        except UnicodeError:
            pass

        try:
            data['contributors'] = list(self.obj.creators + self.obj.contributors)
        except:
            pass
        path = self.obj.getPhysicalPath()
        data['parent_folder'] = '/'.join(path[:-1])
        site_path = api.portal.get().getPhysicalPath()
        if len(path) > (len(site_path) + 1):
            data['immediate_folder'] = path[len(site_path):][0]
        else:
            data['immediate_folder'] = '/'
        return data


class QueryAssembler(query.QueryAssembler):
    def __call__(self, dquery):
        dquery['trashed'] = False
        query = super(QueryAssembler, self).__call__(dquery)
        # take into account views, likes and custom weighting
        try:
            searchq = dquery.get('SearchableText', '')
            if isinstance(searchq, dict):
                searchq = searchq.get('query', '')
            searchq = searchq.lower().strip('*')
            query = {
                "function_score": {
                    "query": query,
                    # "boost_mode": "sum",  # add score and modified score,
                    "script_score": {
                        "lang": "native",
                        'params': {
                            'search': searchq
                        },
                        "script": "castlepopularity"
                    }
                }
            }
        except KeyError:
            pass
        return query



from plone.app.contenttypes import _
from plone.app.contenttypes.behaviors.leadimage import ILeadImage
from plone.autoform.interfaces import IFormFieldProvider
from plone.dexterity.interfaces import IDexterityContent
from plone.namedfile import field as namedfile
from zope import schema
from zope.component import adapter
from zope.interface import implementer
from zope.interface import provider


@provider(IFormFieldProvider)
class IRequiredLeadImage(ILeadImage):

    image = namedfile.NamedBlobImage(
        title=_(u'label_leadimage', default=u'Lead Image'),
        description=_(u'help_leadimage', default=u''),
        required=True
    )

    image_caption = schema.TextLine(
        title=_(u'label_leadimage_caption', default=u'Lead Image Caption'),
        description=_(u'help_leadimage_caption', default=u''),
        required=False,
    )


@implementer(IRequiredLeadImage)
@adapter(IDexterityContent)
class RequiredLeadImage(object):

    def __init__(self, context):
        self.context = context



from Acquisition import aq_parent
from base64 import b64decode
from castle.cms.behaviors.leadimage import IRequiredLeadImage
from castle.cms.interfaces import ICastleLayer
from castle.cms.interfaces import IReCaptchaWidget
from castle.cms.interfaces import IReferenceNamedImage
from castle.cms.tiles.views import getTileViews
from plone.app.contenttypes.behaviors.leadimage import ILeadImage
from plone.app.uuid.utils import uuidToObject
from plone.app.widgets.base import InputWidget as BaseInputWidget
from plone.app.widgets.base import TextareaWidget as BaseTextareaWidget
from plone.app.z3cform.widget import AjaxSelectWidget as pz3c_AjaxSelectWidget
from plone.app.z3cform.widget import QueryStringWidget as BaseQueryStringWidget
from plone.app.z3cform.widget import RelatedItemsWidget as BaseRelatedItemsWidget
from plone.app.z3cform.widget import SelectWidget as pz3c_SelectWidget
from plone.app.z3cform.widget import BaseWidget
from plone.formwidget.namedfile.converter import NamedDataConverter
from plone.formwidget.namedfile.interfaces import INamedImageWidget
from plone.formwidget.namedfile.widget import NamedImageWidget as BaseNamedImageWidget
from plone.namedfile.file import NamedBlobImage
from plone.namedfile.interfaces import INamedImageField
from plone.namedfile.utils import safe_basename
from plone.registry.interfaces import IRegistry
from z3c.form.browser import text
from z3c.form.browser.checkbox import SingleCheckBoxWidget
from z3c.form.browser.select import SelectWidget as z3cform_SelectWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.interfaces import ITextWidget
from z3c.form.util import getSpecification
from zope.component import adapter
from zope.component import adapts
from zope.component import getUtility
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.interface import implements
from zope.interface import implementsOnly
from zope.schema.interfaces import IField
from ZPublisher.HTTPRequest import FileUpload

import json
import z3c.form.browser.textarea
import z3c.form.interfaces
import z3c.form.widget


class MultiSelectWidget(pz3c_SelectWidget):
    multiple = True


class QueryStringWidget(BaseQueryStringWidget):
    """
    okay, maybe this sucks, I'm not sure.
    but we're sort of customizing how the widget works with our tiles here...
    """

    def _base_args(self):
        args = super(QueryStringWidget, self)._base_args()
        args['pattern_options']['showPreviews'] = False
        return args


class RelatedItemsWidget(BaseRelatedItemsWidget):
    initialPath = None
    base_criteria = []

    def _base_args(self):
        args = super(RelatedItemsWidget, self)._base_args()
        args['pattern_options']['width'] = ''
        args['pattern_options']['initialPath'] = self.initialPath
        base_criteria = self.base_criteria[:]
        args['pattern_options']['baseCriteria'] = base_criteria
        return args


class RelatedItemWidget(RelatedItemsWidget):

    def _base_args(self):
        args = super(RelatedItemWidget, self)._base_args()
        args['pattern_options']['maximumSelectionSize'] = 1
        return args


class ImageRelatedItemsWidget(RelatedItemsWidget):

    initialPath = '/image-repository'

    def _base_args(self):
        args = super(ImageRelatedItemsWidget, self)._base_args()
        args['pattern_options']['selectableTypes'] = ['Image']
        args['pattern_options']['baseCriteria'].append({
            'i': 'portal_type',
            'o': 'plone.app.querystring.operation.selection.any',
            'v': ['Image', 'Folder']
        })
        return args


class ImageRelatedItemWidget(RelatedItemsWidget):

    initialPath = '/image-repository'

    def _base_args(self):
        args = super(ImageRelatedItemWidget, self)._base_args()
        args['pattern_options']['maximumSelectionSize'] = 1
        args['pattern_options']['selectableTypes'] = ['Image']
        args['pattern_options']['baseCriteria'].append({
            'i': 'portal_type',
            'o': 'plone.app.querystring.operation.selection.any',
            'v': ['Image', 'Folder']
        })
        return args


class FileRelatedItemsWidget(RelatedItemsWidget):

    initialPath = '/file-repository'

    def _base_args(self):
        args = super(FileRelatedItemsWidget, self)._base_args()
        args['pattern_options']['selectableTypes'] = ['File', 'Audio', 'Video']
        args['pattern_options']['baseCriteria'] = [{
            'i': 'portal_type',
            'o': 'plone.app.querystring.operation.selection.any',
            'v': ['File', 'Audio', 'Video', 'Folder']
        }]
        return args


class VideoRelatedItemsWidget(RelatedItemsWidget):

    initialPath = '/video-repository'

    def _base_args(self):
        args = super(VideoRelatedItemsWidget, self)._base_args()
        args['pattern_options']['selectableTypes'] = ['Video']
        args['pattern_options']['baseCriteria'] = [{
            'i': 'portal_type',
            'o': 'plone.app.querystring.operation.selection.any',
            'v': ['Video', 'Folder']
        }]
        return args


class AudioRelatedItemsWidget(RelatedItemsWidget):

    initialPath = '/audio-repository'

    def _base_args(self):
        args = super(AudioRelatedItemsWidget, self)._base_args()
        args['pattern_options']['selectableTypes'] = ['Audio']
        args['pattern_options']['baseCriteria'] = [{
            'i': 'portal_type',
            'o': 'plone.app.querystring.operation.selection.any',
            'v': ['Audio', 'Folder']
        }]
        return args


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def QueryFieldWidget(field, request):
    return z3c.form.widget.FieldWidget(field, QueryStringWidget(request))


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def RelatedItemsFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, RelatedItemsWidget(request))
    widget.vocabulary = 'plone.app.vocabularies.Catalog'
    return widget


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def RelatedItemFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, RelatedItemWidget(request))
    widget.vocabulary = 'plone.app.vocabularies.Catalog'
    return widget


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def ImageRelatedItemsFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field,
                                         ImageRelatedItemsWidget(request))
    widget.vocabulary = 'plone.app.vocabularies.Catalog'
    return widget


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def ImageRelatedItemFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field,
                                         ImageRelatedItemWidget(request))
    widget.vocabulary = 'plone.app.vocabularies.Catalog'
    return widget


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def FileRelatedItemsFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field,
                                         FileRelatedItemsWidget(request))
    widget.vocabulary = 'plone.app.vocabularies.Catalog'
    return widget


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def VideoRelatedItemsFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field,
                                         VideoRelatedItemsWidget(request))
    widget.vocabulary = 'plone.app.vocabularies.Catalog'
    return widget


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def AudioRelatedItemsFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field,
                                         AudioRelatedItemsWidget(request))
    widget.vocabulary = 'plone.app.vocabularies.Catalog'
    return widget


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def SelectFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field,
                                         pz3c_SelectWidget(request))
    return widget


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def MultiSelectFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field,
                                         MultiSelectWidget(request))
    return widget


class IJSONListWidget(ITextWidget):
    """Marker interface for the Select2Widget."""


class JsonListWidget(BaseWidget):
    """Ajax select widget for z3c.form."""

    _base = BaseInputWidget

    implementsOnly(IJSONListWidget)

    pattern = 'mapselect'
    pattern_options = BaseWidget.pattern_options.copy()

    def _base_args(self):
        args = super(JsonListWidget, self)._base_args()
        args['name'] = self.name
        args['value'] = (self.request.get(self.name,
                                          self.value) or u'[]').strip()
        return args


class IMapMarkersWidget(IJSONListWidget):
    """Marker interface for the Select2Widget."""


class MapMarkersWidget(JsonListWidget):
    implementsOnly(IMapMarkersWidget)

    pattern = 'mapselect'
    pattern_options = JsonListWidget.pattern_options.copy()

    def _base_args(self):
        args = super(MapMarkersWidget, self)._base_args()
        args.setdefault('pattern_options')
        args['pattern_options']['type'] = 'markers'
        return args


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def MapMarkersFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, MapMarkersWidget(request))
    return widget


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def UseQueryWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, SingleCheckBoxWidget(request))
    widget.addClass('pat-castledynamicform')
    return widget


class MapPointWidget(BaseWidget):
    """Ajax select widget for z3c.form."""

    _base = BaseInputWidget

    implementsOnly(IMapMarkersWidget)

    klass = style = title = lang = onclick = ondblclick = onmousedown = ''
    onmouseup = onmouseover = onmousemove = onmouseout = onkeypress = ''
    onkeydown = onkeyup = ''

    pattern = 'mapselect'
    pattern_options = BaseWidget.pattern_options.copy()

    def _base_args(self):
        args = super(MapPointWidget, self)._base_args()
        args['name'] = self.name
        args['value'] = (self.request.get(self.name,
                                          self.value) or u'{}').strip()

        args.setdefault('pattern_options', {})
        args['pattern_options']['type'] = 'point'
        return args


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def MapPointFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, MapPointWidget(request))
    return widget


class MapPointsWidget(MapPointWidget):
    """Ajax select widget for z3c.form."""

    _base = BaseInputWidget

    implementsOnly(IMapMarkersWidget)

    pattern = 'mapselect'
    pattern_options = BaseWidget.pattern_options.copy()

    def _base_args(self):
        args = super(MapPointsWidget, self)._base_args()
        args['value'] = (self.request.get(self.name,
                                          self.value) or u'[]').strip()
        args['pattern_options']['type'] = 'points'
        return args


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def MapPointsFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, MapPointsWidget(request))
    return widget


class PreviewSelectWidget(pz3c_SelectWidget):

    previews = {}
    tile_name = None
    pattern = 'previewselect'

    def _base_args(self):
        args = super(PreviewSelectWidget, self)._base_args()
        args.setdefault('pattern_options', {
        })
        previews = self.previews.copy()
        if self.tile_name:
            for view in getTileViews(aq_parent(self.context), self.request, self.tile_name):
                previews[view.name] = view.preview
        args['pattern_options']['previews'] = previews
        return args


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def PreviewSelectFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, PreviewSelectWidget(request))
    return widget


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def NavigationTypeWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, z3cform_SelectWidget(request))
    widget.addClass('pat-castledynamicform')
    return widget


class ReCaptchaWidget(text.TextWidget):
    maxlength = 7
    size = 8

    implementsOnly(IReCaptchaWidget)

    def public_key(self):
        registry = getUtility(IRegistry)
        return registry.get('castle.recaptcha_public_key', '')


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def ReCaptchaFieldWidget(field, request):
    """IFieldWidget factory for CaptchaWidget."""
    return z3c.form.widget.FieldWidget(field, ReCaptchaWidget(request))


class ITinyMCETextWidget(ITextWidget):
    """Marker interface for the Select2Widget."""


class TinyMCETextWidget(BaseWidget):
    """Ajax select widget for z3c.form."""

    _base = BaseTextareaWidget

    implementsOnly(ITinyMCETextWidget)

    pattern = 'tinymce'
    pattern_options = BaseWidget.pattern_options.copy()

    def _base_args(self):
        args = super(TinyMCETextWidget, self)._base_args()
        args['name'] = self.name
        args['value'] = self.request.get(self.name, self.value) or ''
        return args


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def TinyMCETextFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, TinyMCETextWidget(request))
    return widget


class AjaxSelectWidget(pz3c_AjaxSelectWidget):

    def _base_args(self):
        args = super(AjaxSelectWidget, self)._base_args()
        vurl = args['pattern_options']['vocabularyUrl']
        if vurl.count('@@') == 2:
            # badly generated url
            split = vurl.split('@@')
            args['pattern_options']['vocabularyUrl'] = '%s/@@%s' % (
                split[0], split[-1])
        return args


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def AjaxSelectFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, AjaxSelectWidget(request))
    return widget


class IFocalNamedImageWidget(INamedImageWidget):
    pass


class FocalNamedImageWidget(BaseNamedImageWidget):
    """A widget for a named file object
    """
    implements(IFocalNamedImageWidget)

    def get_image_options(self):
        download_url = self.download_url
        contentType = getattr(self.value, 'contentType', None)
        if contentType and 'gif' in contentType:
            contentType = 'image/png'
            download_url = '%s/@@download-as-png' % self.context.absolute_url()

        width = height = 0
        focal_point = [0, 0]
        try:
            width, height = self.value.getImageSize()
            try:
                focal_point = self.value.focal_point
            except:
                focal_point = self.context._image_focal_point
        except:
            if self.value:
                try:
                    focal_point = [width / 2, height / 2]
                except:
                    pass
        try:
            fct = self.file_content_type
        except:
            fct = None
        try:
            icon = self.file_icon
        except:
            icon = None
        return {
            'exists': self.value is not None,
            'download_url': download_url,
            'filename': self.filename,
            'content_type': contentType,
            'icon': icon,
            'thumb_width': self.thumb_width,
            'file_size': self.file_size,
            'doc_type': fct,
            'width': width,
            'height': height,
            'focal_point': focal_point
        }

    def get_reference_options(self):
        download_url = self.download_url
        if isinstance(self.value, basestring) and self.value.startswith('reference:'):
            reference = self.value.replace('reference:', '')
        else:
            reference = self.value.reference
        return {
            'reference': reference,
            'exists': False,
            'download_url': download_url,
            'filename': self.filename
        }

    @property
    def pattern_options(self):
        result = {
            'reference': None,
            'id': self.id,
            'title': self.title,
            'required': self.required,
            'allow_nochange': self.allow_nochange,
            'name': self.name,
            'disabled': self.disabled,
            'maxlength': self.maxlength
        }
        is_string = isinstance(self.value, basestring)
        if (IReferenceNamedImage.providedBy(self.value) or
                (is_string and self.value.startswith('reference:'))):
            result.update(self.get_reference_options())
        else:
            result.update(self.get_image_options())
        return json.dumps(result)


@adapter(getSpecification(ILeadImage['image']), ICastleLayer)
@implementer(IFieldWidget)
def LeadImageFocalNamedImageFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, FocalNamedImageWidget(request))
    return widget


@adapter(getSpecification(IRequiredLeadImage['image']), ICastleLayer)
@implementer(IFieldWidget)
def RequiredLeadImageFocalNamedImageFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, FocalNamedImageWidget(request))
    return widget


@adapter(INamedImageField, ICastleLayer)
@implementer(IFieldWidget)
def FocalNamedImageFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, FocalNamedImageWidget(request))
    return widget


class FocalNamedImageDataConverter(NamedDataConverter):
    """Converts from a file-upload to a NamedFile variant.
    """
    adapts(INamedImageField, IFocalNamedImageWidget)

    def toWidgetValue(self, value):
        return value

    def getImage(self, value):
        widget = self.widget
        req = widget.request

        if isinstance(value, FileUpload):

            filename = safe_basename(value.filename)

            if filename is not None and not isinstance(filename, unicode):
                # Work-around for
                # https://bugs.launchpad.net/zope2/+bug/499696
                filename = filename.decode('utf-8')

            value.seek(0)
            data = value.read()
            if data or filename:
                value = self.field._type(data=data, filename=filename)
            else:
                return self.get_missing_image()

        else:
            args = {}
            if req.get(widget.name + '.filename'):
                filename = req.get(widget.name + '.filename')
                if type(filename) in (list, set, tuple) and filename:
                    filename = filename[0]
                if not isinstance(filename, unicode):
                    filename = filename.decode('utf8')
                args['filename'] = filename
            if isinstance(value, basestring):
                if value.startswith('reference:'):
                    reference = value.replace('reference:', '')
                    obj = uuidToObject(reference)
                    if obj:
                        try:
                            try:
                                ct = obj.image.contentType
                            except AttributeError:
                                ct = None
                            value = NamedBlobImage(contentType=ct)
                            alsoProvides(value, IReferenceNamedImage)
                            value.reference = reference
                        except AttributeError:
                            pass
                elif value.startswith('data:'):
                    value = b64decode(value.split(',')[-1])
                    value = self.field._type(data=value, **args)
            else:
                try:
                    value = self.field._type(data=str(value), **args)
                except UnicodeEncodeError:
                    value = self.field._type(data=value.encode('utf8'), **args)

        return value

    def get_missing_image(self):
        if hasattr(self.widget.context, self.field.__name__):
            return getattr(self.widget.context, self.field.__name__)
        return self.field.missing_value

    def toFieldValue(self, value):
        widget = self.widget
        req = widget.request
        action = req.get(widget.name + '.action')

        if action == 'remove':
            return self.field.missing_value

        if value is None or value == '':
            return self.get_missing_image()

        if action not in ('nochange',):
            value = self.getImage(value)

        if not value:
            # another attempt to safe guard
            return self.get_missing_image()

        if not IReferenceNamedImage.providedBy(value):
            # set focal point data
            try:
                fp = [
                    float(req.form.get(widget.name + '.focalX')),
                    float(req.form.get(widget.name + '.focalY'))
                ]
                value.focal_point = fp
                self.widget.context._image_focal_point = fp  # backup in case other doesn't save
            except:
                pass
        return value


class IFocalPointSelectWidget(ITextWidget):
    """Marker interface for the Select2Widget."""


class FocalPointSelectWidget(BaseWidget):
    """Widget to select focal point"""

    _base = BaseInputWidget

    implementsOnly(IFocalPointSelectWidget)

    pattern = 'focalpointselect'
    pattern_options = BaseWidget.pattern_options.copy()

    def _base_args(self):
        args = super(FocalPointSelectWidget, self)._base_args()
        args['name'] = self.name
        args['value'] = (self.request.get(self.name,
                                          self.value) or u'').strip()
        return args


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def FocalPointSelectFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, FocalPointSelectWidget(request))
    return widget


class ITOCWidget(IJSONListWidget):
    """Marker interface for the Select2Widget."""


class TOCWidget(JsonListWidget):
    implementsOnly(IMapMarkersWidget)

    pattern = 'toccreator'
    pattern_options = JsonListWidget.pattern_options.copy()


@adapter(IField, ICastleLayer)
@implementer(IFieldWidget)
def TOCFieldWidget(field, request):
    widget = z3c.form.widget.FieldWidget(field, TOCWidget(request))
    return widget



from Products.Five import BrowserView
from castle.cms.lockout import LockoutManager


class LockedOutView(BrowserView):

    def __call__(self):
        manager = LockoutManager(self.context, self.request.get('username'))
        self.attempts = manager.get_attempts_this_window()
        self.manager = manager
        return self.index()


class DisabledUserView(BrowserView):

    def __call__(self):
        return self.index()



from castle.cms.tiles.base import DisplayTypeTileMixin
from castle.cms.tiles.content import ContentTile
from castle.cms.tiles.content import IContentTileSchema
from castle.cms.tiles.views import BaseTileView
from castle.cms.tiles.views import TileViewsSource
from castle.cms.widgets import FocalPointSelectFieldWidget
from castle.cms.widgets import PreviewSelectFieldWidget
from plone.autoform import directives as form
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from z3c.form.browser.checkbox import CheckBoxFieldWidget
from zope import schema
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


DISPLAY_TYPE_KEY = 'existing'


class BasicView(BaseTileView):
    name = 'basic'
    preview = ''
    order = 0
    index = ViewPageTemplateFile('templates/existing/basic.pt')
    tile_name = DISPLAY_TYPE_KEY


class BackgroundImageView(BaseTileView):
    name = 'backgroundimage'
    preview = ''
    order = 1
    index = ViewPageTemplateFile('templates/existing/backgroundimage.pt')
    tile_name = DISPLAY_TYPE_KEY


class ExistingTile(ContentTile, DisplayTypeTileMixin):

    display_type_name = DISPLAY_TYPE_KEY
    display_type_default = 'basic'
    display_type_fallback_view = BasicView

    def joiner(self, val, joiner=', '):
        if not val:
            return ''
        if type(val) in (set, list, tuple):
            val = joiner.join(val)
        if isinstance(val, basestring) and '\n' in val:
            val = joiner.join(val.splitlines())
        return val


class IExistingTileSchema(IContentTileSchema):

    form.order_before(title='use_query')
    title = schema.TextLine(
        title=u'Title heading',
        description=u'Will be omitted if blank',
        required=False,
        default=u'')

    form.widget('display_fields', CheckBoxFieldWidget)
    display_fields = schema.Tuple(
        title=u'Display fields',
        description=u'Fields that should show from the content',
        default=(
            'title',
            'image',
            'description'
        ),
        value_type=schema.Choice(
            vocabulary=SimpleVocabulary([
                SimpleTerm('title', 'title', u'Title'),
                SimpleTerm('image', 'image', u'Image'),
                SimpleTerm('description', 'description', u'Overview/Summary'),
                SimpleTerm('date', 'date', u'Date'),
            ])
        )
    )

    truncate_count = schema.Int(
        title=u'Overview/Summary Truncate',
        description=u'Number of words to truncate the overview/summary to',
        required=False,
        default=18
    )

    image_display_type = schema.Choice(
        title=u'Image display type',
        description=u'Does not apply to all display types',
        required=True,
        default=u'landscape',
        vocabulary=SimpleVocabulary([
            SimpleTerm('landscape', 'landscape', u'Landscape'),
            SimpleTerm('portrait', 'portrait', u'Portrait'),
            SimpleTerm('square', 'square', u'Square')
        ])
    )

    display_type = schema.Choice(
        title=u"Display Type",
        vocabulary=SimpleVocabulary([
            SimpleTerm('basic', 'basic', u'Basic'),
            SimpleTerm('backgroundimage', 'backgroundimage', u'Background Image')
        ]),
        default='basic'
    )

    form.widget('display_type', PreviewSelectFieldWidget,
                tile_name=DISPLAY_TYPE_KEY)
    display_type = schema.Choice(
        title=u"Display Type",
        source=TileViewsSource(DISPLAY_TYPE_KEY),
        default='default'
    )

    form.widget(override_focal_point=FocalPointSelectFieldWidget)
    override_focal_point = schema.Text(
        title=u'Override Focal point',
        default=u'',
        required=False
    )



from Acquisition import aq_base
from castle.cms.browser.utils import Utils
from castle.cms.interfaces import IAudio
from castle.cms.interfaces import ILDData
from castle.cms.interfaces import IVideo
from plone import api
from plone.dexterity.interfaces import IDexterityContent
from plone.registry.interfaces import IRegistry
from Products.CMFCore.interfaces import ISiteRoot
from Products.CMFPlone.browser.syndication.adapters import DexterityItem
from Products.CMFPlone.browser.syndication.adapters import FolderFeed
from Products.CMFPlone.interfaces import ISiteSchema
from Products.CMFPlone.interfaces.syndication import IFeedItem
from Products.CMFPlone.utils import getSiteLogo
from ZODB.POSException import POSKeyError
from zope.component import adapts
from zope.component import getUtility
from zope.component import queryMultiAdapter
from zope.globalrequest import getRequest
from zope.interface import implements

import json


def format_date(dt):
    try:
        return dt.ISO8601()
    except:
        return ''


class NoFileDexterityItem(DexterityItem):

    def __init__(self, context, feed):
        super(DexterityItem, self).__init__(context, feed)
        self.dexterity = IDexterityContent.providedBy(context)
        self.file = None


class LDData(object):
    implements(ILDData)
    adapts(IDexterityContent)

    type_ = 'WebPage'

    view_type = False

    def __init__(self, context):
        self.site = api.portal.get()
        feed = FolderFeed(self.site)
        self.item = None
        try:
            self.item = queryMultiAdapter((context, feed), IFeedItem)
        except POSKeyError:
            pass
        if self.item is None:
            self.item = NoFileDexterityItem(context, feed)
        self.context = context
        self.registry = getUtility(IRegistry)

    def get_data(self):
        data = {
            "@context": "http://schema.org",
            "@type": self.type_,
            "name": self.item.title,
            "headline": self.item.title,
            "url": self.item.link,
            "description": self.item.description,
            "datePublished": format_date(self.item.published),
            "dateModified": format_date(self.item.modified),
        }

        if getattr(aq_base(self.context), 'image', None) is not None:
            data['image'] = {
                "@type": "ImageObject",
                "url": "%s/@@images/image" % self.item.base_url,
            }
            data['thumbnailUrl'] = "%s/@@images/image/mini" % self.item.base_url

        registry = getUtility(IRegistry)
        view_about = registry.get('plone.allow_anon_views_about', False)
        if view_about:
            try:
                author = self.context.creators[0]
                data['author'] = {
                    "@type": "Person",
                    "name": author
                }
            except:
                pass
        return data


class LDAudioData(LDData):
    adapts(IAudio)

    _type = 'AudioObject'

    def get_data(self):
        data = super(LDAudioData, self).get_data()
        data['contentUrl'] = self.item.base_url
        try:
            data['caption'] = self.context.transcript.output
        except:
            pass
        return data


class LDVideoData(LDAudioData):
    adapts(IVideo)

    _type = 'VideoObject'


class LDSiteData(object):
    implements(ILDData)
    adapts(ISiteRoot)

    def __init__(self, context):
        self.context = context

    def get_data(self):
        registry = getUtility(IRegistry)
        btype = registry.get('castle.business_type', None)
        bname = registry.get('castle.business_name', None)
        utils = Utils(self.context, getRequest())

        if not btype and bname:
            # return subset that uses site name, org type
            registry = getUtility(IRegistry)
            site_settings = registry.forInterface(ISiteSchema,
                                                  prefix="plone",
                                                  check=False)
            return {
                "@context": "http://schema.org",
                "@type": 'Organization',
                "name": site_settings.site_title,
                "url": utils.get_public_url(),
                "logo": {
                    "@type": "ImageObject",
                    "url": getSiteLogo()
                }
            }

        data = {
            "@context": "http://schema.org",
            "@type": btype,
            "name": bname,
            "url": utils.get_public_url(),
            'telephone': registry.get('castle.business_telephone', ''),
            "logo": {
                "@type": "ImageObject",
                "url": getSiteLogo()
            }
        }
        address = registry.get('castle.business_address', None)
        if address:
            data['address'] = {
                '@type': 'PostalAddress',
                'streetAddress': address,
                'addressLocality': registry.get('castle.business_city', ''),
                'addressRegion': registry.get('castle.business_state', ''),
                'postalCode': registry.get('castle.business_postal_code', ''),
                'addressCountry': registry.get('castle.business_country', '')
            }
        coordinates = registry.get('castle.business_coordinates', '')
        if coordinates:
            try:
                coordinates = json.loads(coordinates)
            except:
                coordinates = None
            if coordinates:
                data['geo'] = {
                    '@type': 'GeoCoordinates',
                    'latitude': coordinates.get('lat', ''),
                    'longitude': coordinates.get('lng', '')
                }
        days = registry.get('castle.business_days_of_week', [])
        hours = []
        if days:
            hours.append({
                '@type': "OpeningHoursSpecification",
                'dayOfWeek': days,
                'opens': registry.get('castle.business_opens', ''),
                'closes': registry.get('castle.business_closes', ''),
            })
        for special in registry.get('castle.business_special_hours', []) or []:
            if special.count('|') != 2:
                continue
            day, opens, closes = special.split('|')
            hours.append({
                '@type': "OpeningHoursSpecification",
                'dayOfWeek': day,
                'opens': opens,
                'closes': closes
            })
        if hours:
            data['openingHoursSpecification'] = hours
        menu = registry.get('castle.business_menu_link', None)
        if menu:
            data["menu"] = menu
        if registry.get('castle.business_accepts_reservations', False):
            data["acceptsReservations"] = "True"

        try:
            data.update(json.loads(
                registry.get('castle.business_additional_configuration', '{}')))
        except:
            pass
        return data



# -*- coding: utf-8 -*-
from castle.cms.behaviors.order import AvailableOrderSource
from castle.cms.behaviors.order import FolderOrder
from castle.cms.ordering.reversed import ReversedOrdering
from castle.cms.testing import CASTLE_PLONE_INTEGRATION_TESTING
from plone import api
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME

import unittest


class TestOrdering(unittest.TestCase):

    layer = CASTLE_PLONE_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']

        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ('Member', 'Manager'))

    def test_available_order_vocab(self):
        folder = api.content.create(type='Folder', title='Foobar', container=self.portal)
        self.assertEquals(len(AvailableOrderSource()(folder)), 3)

    def test_getting_value(self):
        folder = api.content.create(type='Folder', title='Foobar', container=self.portal)
        self.assertEquals(folder.folder_order, '')

    def test_setting_value(self):
        folder = api.content.create(type='Folder', title='Foobar', container=self.portal)
        order = FolderOrder(folder)
        order.folder_order = 'reversed'
        self.assertEquals(type(folder.getOrdering()), ReversedOrdering)


class TestReversedOrdering(unittest.TestCase):
    layer = CASTLE_PLONE_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']

        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ('Member', 'Manager'))

    def test_adding_content_adds_to_beginning(self):
        folder = api.content.create(type='Folder', title='Foobar', container=self.portal)
        order = FolderOrder(folder)
        order.folder_order = 'reversed'
        api.content.create(type='Document', id='one', container=folder)
        api.content.create(type='Document', id='two', container=folder)
        self.assertEquals(folder.objectIds(), ['two', 'one'])

    def test_move_objects(self):
        folder = api.content.create(type='Folder', title='Foobar', container=self.portal)
        order = FolderOrder(folder)
        order.folder_order = 'reversed'
        api.content.create(type='Document', id='one', container=folder)
        api.content.create(type='Document', id='two', container=folder)
        api.content.create(type='Document', id='three', container=folder)

        # order should be 3,2,1 right now
        self.assertEquals(folder.objectIds(), ['three', 'two', 'one'])

        # move 3 to end
        folder.moveObjectsByDelta(('three',), 2)

        self.assertEquals(folder.objectIds(), ['two', 'one', 'three'])

    def test_move_to_bottom(self):
        folder = api.content.create(type='Folder', title='Foobar', container=self.portal)
        order = FolderOrder(folder)
        order.folder_order = 'reversed'
        api.content.create(type='Document', id='one', container=folder)
        api.content.create(type='Document', id='two', container=folder)
        api.content.create(type='Document', id='three', container=folder)

        folder.moveObjectsToBottom(('three',))
        self.assertEquals(folder.objectIds(), ['two', 'one', 'three'])

    def test_move_to_top(self):
        folder = api.content.create(type='Folder', title='Foobar', container=self.portal)
        order = FolderOrder(folder)
        order.folder_order = 'reversed'
        api.content.create(type='Document', id='one', container=folder)
        api.content.create(type='Document', id='two', container=folder)
        api.content.create(type='Document', id='three', container=folder)

        folder.moveObjectsToTop(('one',))
        self.assertEquals(folder.objectIds(), ['one', 'three', 'two'])



from BTrees.OOBTree import OOBTree
from castle.cms import archival
from castle.cms.constants import CRAWLED_DATA_KEY
from castle.cms.constants import CRAWLED_SITE_ES_DOC_TYPE
from castle.cms.cron.utils import login_as_admin
from castle.cms.cron.utils import setup_site
from castle.cms.cron.utils import spoof_request
from castle.cms.files import aws
from castle.cms.interfaces import ICrawlerConfiguration
from collective.elasticsearch.es import ElasticSearchCatalog
from collective.elasticsearch.interfaces import IMappingProvider
from DateTime import DateTime
from elasticsearch import NotFoundError
from lxml import etree
from lxml import html
from persistent.dict import PersistentDict
from plone import api
from plone.registry.interfaces import IRegistry
from Products.CMFPlone.interfaces.siteroot import IPloneSiteRoot
from Products.CMFPlone.log import logger
from tendo import singleton
from urlparse import urlparse
from zope.annotation.interfaces import IAnnotations
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.globalrequest import getRequest

import requests
import sys
import time
import transaction


CRAWLER_ES_MAPPING = {
    'domain': {
        'type': 'string',
        'index': 'not_analyzed',
        'store': False
    },
    'sitemap': {
        'type': 'string',
        'index': 'not_analyzed',
        'store': False
    },
    'url': {
        'type': 'string',
        'index': 'not_analyzed',
        'store': False
    },
    'image_url': {
        'type': 'string',
        'index': 'not_analyzed',
        'store': False
    }
}


class Crawler(object):

    _meta_properties = {
        'Title': [
            'meta[property="og:title"]',
            'title',
            '.documentFirstHeading',
            'h1'
        ],
        'Description': [
            'meta[property="og:description"]',
        ],
        'created': [
            'meta[name="DC.date.created"]',
            'meta[name="creationDate"]',
            'meta[name="publicationDate"]',
        ],
        'modified': [
            'meta[name="DC.date.modified"]',
            'meta[name="modificationDate"]',
            'meta[name="publicationDate"]',
        ],
        'effective': [
            'meta[name="publicationDate"]',
        ],
        'portal_type': [
            'meta[name="DC.type"]',
            'meta[name="portalType"]',
            'meta[property="og:type"]',
        ],
        'image_url': [
            'meta[property="og:image"]',
        ]
    }
    date_fields = ['created', 'modified', 'effective']
    searchable_text_selector = ','.join([
        '#content p,#content h2,#content h3,#content ul li'
    ])

    def __init__(self, site, settings, es):
        self.site = site
        self.settings = settings
        self.es = es
        self.site._p_jar.sync()
        annotations = IAnnotations(site)
        if CRAWLED_DATA_KEY not in annotations:
            annotations[CRAWLED_DATA_KEY] = OOBTree({
                'tracking': PersistentDict()
            })
            transaction.commit()
        self.data = annotations[CRAWLED_DATA_KEY]

    def crawl_page(self, url):
        logger.info('Indexing ' + url)
        resp = requests.get(url, headers={
            'User-Agent': self.settings.crawler_user_agent
        })
        if resp.status_code == 404 or 'html' not in resp.headers.get('content-type', ''):
            # remove from index
            return False
        dom = html.fromstring(resp.content)
        parsed = urlparse(url)
        data = {
            'url': url,
            'domain': parsed.netloc
        }

        for name, selectors in self._meta_properties.items():
            for selector in selectors:
                result = dom.cssselect(selector)
                if len(result) > 0:
                    result = result[0]
                    if result.attrib.get('content'):
                        data[name] = result.attrib['content']
                        break
                    elif result.text_content().strip():
                        data[name] = result.text.strip()
                        break

        for date_field in self.date_fields:
            val = data.get(date_field)
            if val:
                try:
                    data[date_field] = DateTime(val).ISO8601()
                except:
                    pass

        searchable_text = [
            data.get('Title', ''),
            data.get('Description', '')
        ]
        for el in dom.cssselect(self.searchable_text_selector):
            searchable_text.append(el.text_content())
        data['SearchableText'] = ' '.join(searchable_text)

        return data

    def exists_in_index(self, url):
        try:
            self.es.connection.get(
                index=self.es.index_name,
                doc_type=CRAWLED_SITE_ES_DOC_TYPE,
                id=url)
            return True
        except NotFoundError:
            return False

    def crawl_archive_url(self, url):
        if self.exists_in_index(url):
            return
        data = self.crawl_page(url)
        if not data:
            return
        data['sitemap'] = 'archives'
        self.es.connection.index(
            index=self.es.index_name,
            doc_type=CRAWLED_SITE_ES_DOC_TYPE,
            id=url,
            body=data
        )

    def crawl_archives(self):
        registry = getUtility(IRegistry)
        base_url = registry.get('castle.aws_s3_base_url', None)

        storage = archival.Storage(self.site)
        urls = []
        for key, archive_data in storage.archives.items():
            # archives do not need to be re-indexed ever.
            # see if the key is in ES, if it is move on
            url = archive_data.get('view_url', archive_data['url'])
            urls.append(aws.swap_url(url, base_url=base_url))

        query = {
            "filtered": {
                "filter": {
                    "term": {
                        "sitemap": 'archives'
                    }
                },
                "query": {"match_all": {}}
            }
        }
        existing_urls = self.get_all_from_es(query)

        for _id in set(urls) - set(existing_urls):
            # pages that have not yet been crawled
            try:
                self.crawl_archive_url(_id)
            except:
                logger.error('Error indexing archive url: ' + _id, exc_info=True)

        for _id in set(existing_urls) - set(urls):
            # pages that have been removed from the archive
            self.delete_from_index(_id)

    def get_all_from_es(self, query):
        _ids = []
        page_size = 700
        result = self.es.connection.search(
            index=self.es.index_name,
            doc_type=CRAWLED_SITE_ES_DOC_TYPE,
            scroll='30s',
            size=page_size,
            fields=[],
            body={
                "query": query
            })
        _ids.extend([r['_id'] for r in result['hits']['hits']])
        scroll_id = result['_scroll_id']
        while scroll_id:
            result = self.es.connection.scroll(
                scroll_id=scroll_id,
                scroll='30s'
            )
            if len(result['hits']['hits']) == 0:
                break
            _ids.extend([r['_id'] for r in result['hits']['hits']])
            scroll_id = result['_scroll_id']
        return _ids

    def clean_removed_pages(self, sitemap, crawled_urls):
        parsed = urlparse(sitemap)
        domain = parsed.netloc
        query = {
            "filtered": {
                "filter": {
                    "term": {
                        "domain": domain
                    }
                },
                "query": {"match_all": {}}
            }
        }
        ids = self.get_all_from_es(query)
        for _id in set(ids) - set(crawled_urls):
            # what's left are pages we don't care about anymore
            self.delete_from_index(_id)

    def delete_from_index(self, url):
        self.es.connection.delete(
            index=self.es.index_name,
            doc_type=CRAWLED_SITE_ES_DOC_TYPE,
            id=url)

    def crawl_site_map(self, sitemap, full=False):
        resp = requests.get(sitemap, headers={
            'User-Agent': self.settings.crawler_user_agent
        })
        if resp.status_code != 200:
            logger.error('Not a valid sitemap response for %s' % sitemap)
            return

        self.site._p_jar.sync()
        if sitemap in self.data['tracking']:
            last_crawled = DateTime(self.data['tracking'][sitemap])
        else:
            last_crawled = DateTime('1999/01/01')

        self.data['tracking'][sitemap] = DateTime().ISO8601().decode('utf8')
        transaction.commit()

        dom = etree.fromstring(resp.content)
        crawled_urls = []
        for url_node in dom.xpath("//*[local-name() = 'url']"):
            loc = url_node.xpath("*[local-name() = 'loc']")
            if loc:
                loc = loc[0].text.strip()
            else:
                loc = None
            url = loc
            crawled_urls.append(url)

            lastmod = url_node.xpath("*[local-name() = 'lastmod']")
            if lastmod:
                lastmod = lastmod[0].text.strip()
            else:
                lastmod = None
            if lastmod:
                lastmod = DateTime(lastmod)
                if not full and lastmod < last_crawled:
                    continue

            if not url:
                continue

            data = self.crawl_page(url)
            if data is False:
                crawled_urls.remove(url)
                try:
                    self.es.connection.delete(
                        index=self.es.index_name,
                        doc_type=CRAWLED_SITE_ES_DOC_TYPE,
                        id=url)
                except NotFoundError:
                    pass
            else:
                data['sitemap'] = sitemap
                self.es.connection.index(
                    index=self.es.index_name,
                    doc_type=CRAWLED_SITE_ES_DOC_TYPE,
                    id=url,
                    body=data
                )
                crawled_urls.append(url)

        self.clean_removed_pages(sitemap, crawled_urls)


def crawl_site(site, full=False):
    registry = getUtility(IRegistry)
    settings = registry.forInterface(ICrawlerConfiguration, prefix='castle')
    if not settings.crawler_active or not settings.crawler_site_maps:
        return False

    catalog = api.portal.get_tool('portal_catalog')
    es = ElasticSearchCatalog(catalog)
    if not es.enabled:
        return False

    # check index type is mapped, create if not
    try:
        es.connection.indices.get_mapping(
            index=es.index_name,
            doc_type=CRAWLED_SITE_ES_DOC_TYPE)
    except NotFoundError:
        # need to add it
        adapter = getMultiAdapter((getRequest(), es), IMappingProvider)
        mapping = adapter()
        mapping['properties'].update(CRAWLER_ES_MAPPING)
        es.connection.indices.put_mapping(
            doc_type=CRAWLED_SITE_ES_DOC_TYPE,
            body=mapping,
            index=es.index_name)

    crawler = Crawler(site, settings, es)

    if settings.crawler_index_archive:
        crawler.crawl_archives()

    for sitemap in settings.crawler_site_maps:
        try:
            crawler.crawl_site_map(sitemap, full)
        except:
            logger.error('Error crawling site map: %s' % sitemap, exc_info=True)
    return True


def run(app):
    singleton.SingleInstance('crawler')

    app = spoof_request(app)  # noqa
    login_as_admin(app)  # noqa

    count = 0

    while True:
        try:
            if 'site-id' in sys.argv:
                siteid = sys.argv['site-id']
                setup_site(app[siteid])
                crawl_site(app[siteid])  # noqa
            else:
                for oid in app.objectIds():  # noqa
                    obj = app[oid]  # noqa
                    if IPloneSiteRoot.providedBy(obj):
                        try:
                            setup_site(obj)
                            obj._p_jar.sync()
                            crawl_site(obj, count % 10 == 0)
                        except:
                            logger.error('Error crawling site %s' % oid, exc_info=True)
        except KeyError:
            pass
        except:
            logger.error('Error setting up crawling', exc_info=True)

        logger.info('Waiting to crawl again')
        time.sleep(10 * 60)
        count += 1


if __name__ == '__main__':
    run(app)  # noqa



from castle.cms.tiles.base import BaseTile
from castle.cms.browser.subscribe import SubscribeForm
from plone.supermodel import model
from plone import api
import requests, re, json

class SubscriptionTile(BaseTile, SubscribeForm):

    def initialize(self):
        portal = api.portal.get()
        self.request.URL = portal.absolute_url() + '/@@subscribe'
        self.label = u'Subscribe to email updates'

        self.update()

    def __init__(self, context, request):
        super(SubscriptionTile, self).__init__(context, request)
        self.initialize()

class ISubscriptionTileSchema(model.Schema):
    pass



# remove bad archives
# cleanup archive ES entries
# rewrite archive links,images to reference base url
from castle.cms import archival
from castle.cms.cron._crawler import Crawler
from castle.cms.cron.utils import login_as_admin
from castle.cms.files import aws
from castle.cms.interfaces import ICrawlerConfiguration
from collective.elasticsearch.es import ElasticSearchCatalog
from lxml.html import fromstring
from lxml.html import tostring
from plone import api
from plone.registry.interfaces import IRegistry
from urlparse import urlparse
from zope.component import getUtility
from zope.component.hooks import setSite

import argparse
import requests
import transaction


parser = argparse.ArgumentParser(
    description='...')
parser.add_argument('--site-id', dest='site_id', default='Plone')
args, _ = parser.parse_known_args()


def fix_urls(storage, dom):
    for Mover in storage.Movers:
        mover = Mover(dom)
        for el in mover.get_elements():
            url = mover.get_url(el)
            if url is None:
                continue
            # check that the url is an s3 url
            parsed = urlparse(url)
            if parsed.netloc != storage.s3_conn.server_name():
                continue

            original = None
            if 'original-url' in el.attrib:
                # need to maintain the original original
                original = el.attrib['original-url']
            mover.modify(el, aws.swap_url(url))
            if original:
                el.attrib['original-url'] = original

def get_key_from_url(url):
    parsed = urlparse(url)
    # parsed url includes bucket so we strip off bucket to get actual key
    return '/'.join(parsed.path.split('/')[2:])


if __name__ == '__main__':
    login_as_admin(app)  # noqa
    site = app[args.site_id]  # noqa
    setSite(site)

    toremove = {}  # uid: path
    catalog = api.portal.get_tool('portal_catalog')
    registry = getUtility(IRegistry)
    crawler_settings = registry.forInterface(ICrawlerConfiguration, prefix='castle')
    es = ElasticSearchCatalog(catalog)
    crawler = Crawler(site, crawler_settings, es)
    storage = archival.Storage(site)
    for key, archive_data in storage.archives.items():
        for url in (archive_data.get('view_url'), archive_data['url']):
            if not url:
                continue
            resp = requests.get(url)
            if 'html' not in resp.headers.get('content-type'):
                continue
            print('processing ' + url)
            dom = fromstring(resp.content)
            prop = dom.cssselect('meta[property="og:url"]')

            fix_urls(storage, dom)
            html = tostring(dom)

            if html == resp.content:
                # no changes, carry out
                continue

            print('saving new version of ' + url)

            path = get_key_from_url(url)

            # now save back
            url = 'https://{host}/{path}'.format(
                host=storage.s3_conn.server_name(),
                path=path)
            key = storage.bucket.get_key(path)
            key.set_contents_from_string(html, headers={
                'Content-Type': 'text/html; charset=utf-8'
            }, replace=True)
            key.make_public()
    app._p_jar.sync()  # noqa
    print('%i needing to be deleted' % len(toremove))
    for uid, content_path in toremove.items():
        data = storage.archives[uid]
        view_url = data.get('view_url')
        if view_url:
            path = get_key_from_url(url)
            storage.bucket.delete_key(path)
        url = data['url']
        path = get_key_from_url(url)
        storage.bucket.delete_key(path)
        del storage.archives[uid]
        del storage.path_to_uid[content_path]

    transaction.commit()



from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManager import setSecurityPolicy
from Products.CMFCore.tests.base.security import OmnipotentUser
from Products.CMFCore.tests.base.security import PermissiveSecurityPolicy
from Products.CMFPlone.interfaces import IPloneSiteRoot
from Testing.makerequest import makerequest
from zope.app.publication.interfaces import BeforeTraverseEvent
from zope.component.hooks import setSite
from zope.event import notify
from zope.globalrequest import setRequest


def spoof_request(app):
    """
    Make REQUEST variable to be available on the Zope application server.

    This allows acquisition to work properly
    """
    _policy = PermissiveSecurityPolicy()
    _oldpolicy = setSecurityPolicy(_policy)  # noqa
    newSecurityManager(None, OmnipotentUser().__of__(app.acl_users))
    return makerequest(app)


def setup_site(site):
    setSite(site)
    site.clearCurrentSkin()
    site.setupCurrentSkin(site.REQUEST)
    notify(BeforeTraverseEvent(site, site.REQUEST))
    setRequest(site.REQUEST)


def login_as_admin(app):
    user = app.acl_users.getUser('admin')
    newSecurityManager(None, user.__of__(app.acl_users))


def get_sites(app):
    for oid in app.objectIds():  # noqa
        obj = app[oid]  # noqa
        if IPloneSiteRoot.providedBy(obj):
            yield obj



from plone import api
from Products.PluggableAuthService.PluggableAuthService import _noroles
from zope.security import checkPermission
from AccessControl.SecurityManagement import noSecurityManager


ORIGINAL_USER_KEY = 'castle.cms.original_user'
COOKIE_NAME = 'impersonate'


def PAS_validate(self, request, auth='', roles=_noroles):
    user = self._old_validate(request, auth, roles)
    if user is None:
        return

    if request.cookies.get(COOKIE_NAME):
        # check if user can manage portal
        replacement_user_id = request.cookies[COOKIE_NAME]

        if replacement_user_id != 'ANONYMOUS':
            # allow all users to impersonate anonymous users
            try:
                if not checkPermission('cmf.ManagePortal', api.portal.get()):
                    return user
            except api.exc.CannotGetPortalError:
                return user

        (accessed, container, name, value) = self._getObjectContext(
            request['PUBLISHED'], request)
        if replacement_user_id == 'ANONYMOUS':
            plugins = self._getOb('plugins')
            replacement_user = self._createAnonymousUser(plugins)
        else:
            replacement_user = api.user.get(replacement_user_id)
            if replacement_user is not None:
                replacement_user = replacement_user.getUser()
        noSecurityManager()
        if self._authorizeUser(replacement_user, accessed, container,
                               name, value, roles):
            request.environ[ORIGINAL_USER_KEY] = user
            # required to use __of__ so user is wrapped in correct aq context
            return replacement_user.__of__(self)
        else:
            # if user doesn't authenticate, return None
            return None

    return user



from castle.cms.interfaces import IReferenceNamedImage
from AccessControl import Unauthorized
from Acquisition import aq_base
from Acquisition import aq_parent
from castle.cms.behaviors.location import ILocation
from castle.cms.interfaces import IHasDefaultImage
from castle.cms.interfaces import ITrashed
from plone import api
from plone.app.contenttypes.interfaces import IFile
from plone.app.contenttypes.interfaces import IImage
from plone.dexterity.interfaces import IDexterityContent
from plone.indexer.decorator import indexer
from plone.uuid.interfaces import IUUID
from ZODB.POSException import POSKeyError


@indexer(IDexterityContent)
def getRawRelatedItems(obj):
    try:
        result = []
        for relation in obj.relatedItems:
            try:
                to_obj = relation.to_object
                if to_obj:
                    uuid = IUUID(to_obj)
                    if uuid:
                        result.append(uuid)
            except AttributeError:
                pass
        return result
    except AttributeError:
        return []


@indexer(ILocation)
def getLocation(obj):
    bdata = ILocation(obj, None)
    if bdata and bdata.location:
        return bdata.location[0]


@indexer(IDexterityContent)
def hasImage(obj):
    if IHasDefaultImage.providedBy(obj):
        return True
    return getattr(aq_base(obj), 'image', None) is not None


@indexer(IDexterityContent)
def image_info(obj):
    try:
        image = obj.image
        if IReferenceNamedImage.providedBy(image):
            data = {'reference': image.reference}
        else:
            width, height = image.getImageSize()
            data = {
                'width': width,
                'height': height,
                'focal_point': [width / 2, height / 2]
            }

            try:
                data['focal_point'] = image.focal_point
            except AttributeError:
                try:
                    data['focal_point'] = obj._image_focal_point
                except:
                    pass
        return data
    except AttributeError:
        pass


@indexer(IDexterityContent)
def getContentTypeContent(obj):
    return 'text/html'


@indexer(IFile)
def getContentTypeFile(obj):
    try:
        return obj.file.original_content_type
    except:
        try:
            return obj.file.contentType
        except:
            pass


@indexer(IImage)
def getContentTypeImage(obj):
    try:
        return obj.image.contentType
    except:
        pass


@indexer(IDexterityContent)
def trashed(obj):
    while obj:
        if ITrashed.providedBy(obj):
            return True
        obj = aq_parent(obj)
    return False


@indexer(IDexterityContent)
def last_modified_by(context):
    try:
        creator = api.user.get_current().getId()
    except:
        try:
            creator = context.Creators()[0]
        except:
            creator = 'admin'

    rt = api.portal.get_tool("portal_repository")

    if rt is None or not rt.isVersionable(context):
        # not versionable; fallback to the creator
        return creator

    try:
        history = rt.getHistoryMetadata(context)
    except Unauthorized:
        history = None
    if not history:
        return creator

    try:
        if not rt.isUpToDate(context):
            return creator
    except POSKeyError:
        return creator

    length = history.getLength(countPurged=False)

    last = history.retrieve(length - 1)
    if not last or type(last) != dict:
        # unexpected version metadata; fallback to the creator
        return creator

    metadata = last.get("metadata")
    if not metadata or type(metadata) != dict:
        # unexpected version metadata; fallback to the creator
        return creator

    sys_metadata = metadata.get("sys_metadata")
    if not sys_metadata or type(sys_metadata) != dict:
        # unexpected version metadata; fallback to the creator
        return creator

    principal = sys_metadata.get("principal")
    if not principal or type(principal) != str:
        # unexpected version metadata; fallback to the creator
        return creator

    return principal



# always want to be able to steal lock
from plone.protect.auto import safeWrite
from plone.locking import lockable
from plone.locking.interfaces import STEALABLE_LOCK


class TTWLockable(lockable.TTWLockable):
    def unlock(self, lock_type=STEALABLE_LOCK, stealable_only=True):
        self.clear_locks()
        locks = self._locks()
        safeWrite(locks)
        safeWrite(self.context)

    def stealable(self, lock_type=STEALABLE_LOCK):
        return True



from lxml.html import tostring
from lxml.html import fromstring
from castle.cms.archival import BaseArchivalTransformer


class ExampleArchivalTransformer(BaseArchivalTransformer):

    def __call__(self, html):
        dom = fromstring(html)
        dom.cssselect('.foo')[0].text = 'bar'
        return tostring(dom)



from ZPublisher.BaseRequest import quote


_protocol_headers = (
    'HTTP_X_FORWARDED_PROTO',
    'HTTP_X_FORWARDED_PROTOCOL',
    'HTTP_X_SCHEME'
)


def _get_protocol(req):
    for header in _protocol_headers:
        if header in req.environ:
            return req.environ[header]
    return 'http'


def traversal(self, client, request, response=None):
    """
    This is a copy of
    Products.SiteAccess.VirtualHostMonster.VirtualHostMonster.__call__
    with tweaks for:
        1. better detect protocol
        2. better detect host name

    with the ability to already map domains, better predicting http vs https
    and host names is mostly all we need here...
    """
    vh_used = 0
    stack = request['TraversalRequestNameStack']
    path = None

    # detect forwarded for host, port, scheme automatically
    protocol = _get_protocol(request)
    host = request['SERVER_URL'].split('://')[1].lower()
    host, _, port = request.environ.get('HTTP_HOST', host).partition(':')
    port = port or None
    if port:
        server_url = '{}://{}:{}'.format(protocol, host, port)
    else:
        server_url = '{}://{}'.format(protocol, host)

    if server_url != request.SERVER_URL:
        request.setServerURL(protocol, host, port)
        stack_path = stack[:]
        stack_path.reverse()
        request['VIRTUAL_URL_PARTS'] = vup = (
            request['SERVER_URL'], quote('/'.join(stack_path)))
        request['VIRTUAL_URL'] = '/'.join(vup)

        # new ACTUAL_URL
        add = (stack_path and
               request['ACTUAL_URL'].endswith('/')) and '/' or ''
        request['ACTUAL_URL'] = request['VIRTUAL_URL'] + add

    while 1:
        if stack and stack[-1] == 'VirtualHostBase':
            vh_used = 1
            stack.pop()
            protocol = stack.pop()
            host = stack.pop()
            if ':' in host:
                host, port = host.split(':')
                request.setServerURL(protocol, host, port)
            else:
                request.setServerURL(protocol, host)
            path = list(stack)

        # Find and convert VirtualHostRoot directive
        # If it is followed by one or more path elements that each
        # start with '_vh_', use them to construct the path to the
        # virtual root.
        vh = -1
        for ii in range(len(stack)):
            if stack[ii] == 'VirtualHostRoot':
                vh_used = 1
                pp = ['']
                at_end = (ii == len(stack) - 1)
                if vh >= 0:
                    for jj in range(vh, ii):
                        pp.insert(1, stack[jj][4:])
                    stack[vh:ii + 1] = ['/'.join(pp), self.id]
                    ii = vh + 1
                elif ii > 0 and stack[ii - 1][:1] == '/':
                    pp = stack[ii - 1].split('/')
                    stack[ii] = self.id
                else:
                    stack[ii] = self.id
                    stack.insert(ii, '/')
                    ii += 1
                path = stack[:ii]
                # If the directive is on top of the stack, go ahead
                # and process it right away.
                if at_end:
                    request.setVirtualRoot(pp)
                    del stack[-2:]
                break
            elif vh < 0 and stack[ii][:4] == '_vh_':
                vh = ii

        if vh_used or not self.have_map:
            if path is not None:
                path.reverse()
                vh_part = ''
                if path and path[0].startswith('/'):
                    vh_part = path.pop(0)[1:]
                if vh_part:
                    request['VIRTUAL_URL_PARTS'] = vup = (
                        request['SERVER_URL'],
                        vh_part, quote('/'.join(path)))
                else:
                    request['VIRTUAL_URL_PARTS'] = vup = (
                        request['SERVER_URL'], quote('/'.join(path)))
                request['VIRTUAL_URL'] = '/'.join(vup)

                # new ACTUAL_URL
                add = (path and
                       request['ACTUAL_URL'].endswith('/')) and '/' or ''
                request['ACTUAL_URL'] = request['VIRTUAL_URL'] + add

            return

        vh_used = 1  # Only retry once.
        # Try to apply the host map if one exists, and if no
        # VirtualHost directives were found.

        if self.sub_map:
            return

        host = request['SERVER_URL'].split('://')[1].lower()
        hostname, port = (host.split(':', 1) + [None])[:2]
        ports = self.fixed_map.get(hostname, 0)
        if not ports:
            get = self.sub_map.get
            while hostname:
                ports = get(hostname, 0)
                if ports:
                    break
                if '.' not in hostname:
                    return
                hostname = hostname.split('.', 1)[1]
        elif ports:
            pp = ports.get(port, 0)
            if pp == 0 and port is not None:
                # Try default port
                pp = ports.get(None, 0)
            if not pp:
                return
            # If there was no explicit VirtualHostRoot, add one at the end
            if pp[0] == '/':
                pp = pp[:]
                pp.insert(1, self.id)
            stack.extend(pp)



from ZServer.medusa import http_server

import App
import logging
import time


logger = logging.getLogger('castle.cms')

try:
    server = App.config.getConfiguration().servers[0]
except:
    server = None


def medua_log_request(medusa_request, bytes):
    """
    we are patching default logging to disable so we can use our better
    logging with the actual request ob...
    """
    pass
http_server.http_request.log = medua_log_request


def log_date_string(when):
    logtime = time.localtime(when)
    return time.strftime('%d/', logtime) + \
           http_server.http_date.monthname[logtime[1]] + \
           time.strftime('/%Y:%H:%M:%S ', logtime) + \
           http_server.tz_for_log


def log_request(request):
    if server is not None:
        try:
            user = request['AUTHENTICATED_USER']
        except (KeyError, TypeError):
            user = None

        user_name = user
        try:
            user_name = user.name
        except AttributeError:
            try:
                user_name = user.getUserName()
            except AttributeError:
                pass

        if not user_name:
            user_name = 'Anonymous User'

        origin = request.environ['REMOTE_ADDR']
        if request.get_header('x-forwarded-for'):
            forwarded = request.get_header('x-forwarded-for')
            forwarded = forwarded.split(',')[-1].strip()
            if forwarded:
                origin = forwarded

        url = request.environ['PATH_INFO']
        if request.environ.get('QUERY_STRING'):
            url += '?' + request.environ.get('QUERY_STRING')
        request_info = '{} {} {}'.format(
            request.REQUEST_METHOD,
            url,
            request.environ['SERVER_PROTOCOL']
        )

        user_agent = request.get_header('user-agent') or ''
        referer = request.get_header('referer') or ''

        resp = request.response
        server.logger.log(
            origin,
            '- %s [%s] "%s" %d %s "%s" "%s"\n' % (
                user_name,
                log_date_string(time.time()),
                request_info,
                resp.getStatus(),
                resp.headers.get('content-length', '0'),
                referer,
                user_agent
                )
            )



from DateTime import DateTime
from copy import deepcopy
from persistent.dict import PersistentDict
from lxml import etree
from lxml.html import tostring
from repoze.xmliter.utils import getHTMLSerializer
from castle.cms import theming
from plone.app.theming.interfaces import THEME_RESOURCE_NAME
from castle.cms.tiles import meta
from plone import api
from plone.tiles.data import ANNOTATIONS_KEY_PREFIX as TILE_ANNOTATIONS_KEY_PREFIX
from plone.tiles.interfaces import IPersistentTile
from plone.tiles.interfaces import ITileDataManager
from plone.tiles.interfaces import ITileType
from plone.uuid.interfaces import IUUIDGenerator
from Products.Five import BrowserView
from zope.annotation.interfaces import IAnnotations
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.component import queryUtility
from zope.event import notify
from zope.interface import alsoProvides
from zope.lifecycleevent import ObjectRemovedEvent
from castle.cms.events import MetaTileEditedEvent
import urllib

import json


class MetaTileManager(BrowserView):

    def __call__(self):
        self.request.response.setHeader('Content-type', 'application/json')
        if self.request.get('action') == 'addtile':
            self.add_tile()
        elif self.request.get('action') == 'deletetile':
            self.delete_tile()
        elif self.request.get('action') == 'moveup':
            self.move_up()
        elif self.request.get('action') == 'movedown':
            self.move_down()
        elif self.request.get('action') == 'mode':
            self.change_mode()
        elif self.request.get('action') == 'copymeta':
            # XXX disabling working copy support
            return self.copy_meta()
        elif self.request.get('action') == 'savecopy':
            return self.save_copy()
        elif self.request.get('action') == 'cancelcopy':
            return self.cancel_copy()
        return self.info()

    def copy_meta(self):
        # XXX disable working copy support
        return json.dumps({
            'newId': self.request.get('metaId'),
            'success': True,
            'locked': False
        })
        _id = self.request.get('metaId')
        copy_id = self.get_working_copy_meta_id()
        annotations = IAnnotations(self.context)
        data = annotations.get(TILE_ANNOTATIONS_KEY_PREFIX + '.' + _id)
        if not data:
            data = PersistentDict()

        if ('locked' in data and
                self.request.get('override', '').lower() not in ('y', 'yes', '1',
                                                                 'true', 't') and
                api.user.get_current().getId() != data['locked']['user']):
            return json.dumps({
                'locked': True,
                'success': False,
                'lock_data': data['locked']
            })

        if TILE_ANNOTATIONS_KEY_PREFIX + '.' + copy_id in annotations:
            # cut out, we're good, resume existing
            return json.dumps({
                'newId': copy_id,
                'success': True,
                'locked': False
            })

        version_key = self.get_working_copy_key()
        tile_mapping = {}
        new_tiles = []
        for tile in data.get('tiles', []):
            # make copies of all the tiles
            tile_id = tile['id']
            copy_tile_id = '{}-{}'.format(tile_id, version_key)
            tile_data = annotations.get(TILE_ANNOTATIONS_KEY_PREFIX + '.' + tile_id)
            if tile_data:
                annotations[TILE_ANNOTATIONS_KEY_PREFIX + '.' + copy_tile_id] = deepcopy(tile_data)
            new_tile_info = deepcopy(tile)
            new_tile_info['id'] = copy_tile_id
            new_tiles.append(new_tile_info)
            tile_mapping[tile_id] = copy_tile_id

        new_data = PersistentDict(dict(data))
        new_data.update({
            'tiles': new_tiles,
            'mapping': tile_mapping
        })
        data.update({
            'locked': {
                'when': DateTime().ISO8601(),
                'user': api.user.get_current().getId()
            }
        })
        annotations[TILE_ANNOTATIONS_KEY_PREFIX + '.' + copy_id] = new_data

        return json.dumps({
            'newId': copy_id,
            'success': True,
            'locked': False
        })

    def save_copy(self):
        # XXX disable working copy support
        return json.dumps({
            'success': True
        })
        annotations = IAnnotations(self.context)
        _id = self.request.get('metaId')
        copy_id = self.get_working_copy_meta_id()
        existing_data = annotations.get(TILE_ANNOTATIONS_KEY_PREFIX + '.' + _id, {})
        new_data = annotations.get(TILE_ANNOTATIONS_KEY_PREFIX + '.' + copy_id)

        existing_tiles = existing_data.get('tiles', [])
        new_tiles = new_data.get('tiles', [])
        new_tile_ids = [t['id'] for t in new_tiles]
        # tile id: copy tile id
        tile_mapping = new_data.get('mapping', {})
        # copy tile id: tile id
        tile_mapping_reversed = {v: k for k, v in tile_mapping.items()}

        # remove deleted tiles
        for existing_tile in existing_tiles:
            tile_id = existing_tile['id']
            copy_tile_id = tile_mapping.get(tile_id, None)
            if copy_tile_id not in new_tile_ids:
                # was removed, so we delete the data associated with it
                if TILE_ANNOTATIONS_KEY_PREFIX + '.' + tile_id in annotations:
                    del annotations[TILE_ANNOTATIONS_KEY_PREFIX + '.' + tile_id]

        # overwrite new tile data
        for copy_tile in new_tiles:
            copy_tile_id = copy_tile['id']
            if copy_tile_id in tile_mapping_reversed:
                tile_id = tile_mapping_reversed[copy_tile_id]
                # move tile data onto correct key
                if TILE_ANNOTATIONS_KEY_PREFIX + '.' + copy_tile_id in annotations:
                    annotations[TILE_ANNOTATIONS_KEY_PREFIX + '.' + tile_id] = \
                        annotations[TILE_ANNOTATIONS_KEY_PREFIX + '.' + copy_tile_id]
                    del annotations[TILE_ANNOTATIONS_KEY_PREFIX + '.' + copy_tile_id]
                    copy_tile['id'] = tile_id

        # now save the meta tile
        annotations[TILE_ANNOTATIONS_KEY_PREFIX + '.' + _id] = \
            annotations[TILE_ANNOTATIONS_KEY_PREFIX + '.' + copy_id]
        del annotations[TILE_ANNOTATIONS_KEY_PREFIX + '.' + copy_id]

        # also, destory any other active edits for this slot
        for key in list(annotations.keys()):
            if key.startswith(TILE_ANNOTATIONS_KEY_PREFIX + '.' + _id + '-copy-'):
                # also delete tiles referenced here.
                if key not in annotations:
                    continue
                data = annotations[key]
                for tile in data.get('tiles', []):
                    # since these are all copied tiles or new tiles, we can remove
                    tile_id = tile['id']
                    if TILE_ANNOTATIONS_KEY_PREFIX + '.' + tile_id in annotations:
                        del annotations[TILE_ANNOTATIONS_KEY_PREFIX + '.' + tile_id]
                del annotations[key]

        notify(MetaTileEditedEvent(self.context, _id))

        return json.dumps({
            'success': True
        })

    def cancel_copy(self):
        annotations = IAnnotations(self.context)

        _id = self.request.get('metaId')
        copy_id = self.get_working_copy_meta_id()
        existing_data = annotations.get(TILE_ANNOTATIONS_KEY_PREFIX + '.' + _id, {})
        new_data = annotations.get(TILE_ANNOTATIONS_KEY_PREFIX + '.' + copy_id, {})

        for tile in new_data.get('tiles', []):
            tile_id = tile['id']
            # since these are all copied tiles or new tiles, we can remove
            if TILE_ANNOTATIONS_KEY_PREFIX + '.' + tile_id in annotations:
                del annotations[TILE_ANNOTATIONS_KEY_PREFIX + '.' + tile_id]
        if TILE_ANNOTATIONS_KEY_PREFIX + '.' + copy_id in annotations:
            del annotations[TILE_ANNOTATIONS_KEY_PREFIX + '.' + copy_id]

        if 'locked' in existing_data:
            del existing_data['locked']

        return json.dumps({
            'success': True
        })

    def get_working_copy_key(self):
        user = api.user.get_current()
        return 'copy-{}'.format(user.getId())

    def get_working_copy_meta_id(self):
        # XXX disabling locking support
        _id = self.request.get('metaId')
        return _id
        return '{}-{}'.format(_id, self.get_working_copy_key())

    def get_tile(self, tile_type, _id):
        tile = getMultiAdapter((self.context, self.request), name=tile_type)
        alsoProvides(tile, IPersistentTile)
        tile.id = _id
        tile.__name__ = tile_type
        return tile

    def get_meta_data_manager(self):
        _id = self.get_working_copy_meta_id()
        meta_tile = meta.getTile(self.context, self.request, _id)
        return ITileDataManager(meta_tile)

    def move_up(self):
        idx = int(self.request.get('idx'))
        if idx == 0:
            return
        meta_data_manager = self.get_meta_data_manager()
        meta_data = meta_data_manager.get()
        tiles = meta_data['tiles']
        tomove = tiles[idx - 1]
        tiles[idx - 1] = tiles[idx]
        tiles[idx] = tomove

        meta_data['tiles'] = tiles
        meta_data_manager.set(meta_data)

    def move_down(self):
        idx = int(self.request.get('idx'))
        meta_data_manager = self.get_meta_data_manager()
        meta_data = meta_data_manager.get()
        tiles = meta_data['tiles']

        if idx >= len(tiles) - 1:
            return

        tomove = tiles[idx + 1]
        tiles[idx + 1] = tiles[idx]
        tiles[idx] = tomove

        meta_data['tiles'] = tiles
        meta_data_manager.set(meta_data)
        pass

    def add_tile(self):
        generator = getUtility(IUUIDGenerator)
        tile_id = generator()
        tile_type = self.request.get('type')
        tile = self.get_tile(tile_type, tile_id)
        data_manager = ITileDataManager(tile)
        data_manager.set({})

        meta_data_manager = self.get_meta_data_manager()
        data = meta_data_manager.get()
        tiles = data.get('tiles', []) or []
        tiles.append({
            'id': tile_id,
            'type': tile_type
            })
        data['tiles'] = tiles
        meta_data_manager.set(data)

    def delete_tile(self):
        meta_data_manager = self.get_meta_data_manager()
        meta_data = meta_data_manager.get()

        idx = int(self.request.get('idx'))
        tiles = meta_data['tiles']
        tile_data = tiles[idx]
        tile_type = tile_data['type']

        tile = self.get_tile(tile_type, tile_data['id'])

        dataManager = ITileDataManager(tile)
        dataManager.delete()

        notify(ObjectRemovedEvent(tile))

        tiles.remove(tile_data)
        meta_data_manager.set(meta_data)

    def change_mode(self):
        meta_data_manager = self.get_meta_data_manager()
        meta_data = meta_data_manager.get()
        meta_data['mode'] = self.request.get('value')
        meta_data_manager.set(meta_data)

    def addTilesData(self, meta_tile):
        tiles = meta_tile.data.get('tiles', [])[:]
        annotations = IAnnotations(meta_tile.context)
        for tile in tiles:
            tile_type = queryUtility(ITileType, name=tile['type'])
            tile['label'] = tile_type.title
            if tile['type'] == 'castle.cms.fragment':
                data = annotations.get(TILE_ANNOTATIONS_KEY_PREFIX + '.' + tile['id'])
                if data and data.get('fragment'):
                    name = data.get('fragment')
                    tile['label'] += ': ' + name.capitalize().replace('-', ' ')
        return tiles

    def info(self):
        portal = api.portal.get()
        _id = self.get_working_copy_meta_id()

        portal_path = '/'.join(portal.getPhysicalPath())
        current_path = '/'.join(self.context.getPhysicalPath())[len(portal_path):] or '/'

        effective_tiles = []
        for effective_tile in meta.getEffectiveTiles(self.context, self.request, _id,
                                                     self.request.form.get('metaId')):
            effective_tiles.append({
                'path': '/'.join(effective_tile.context.getPhysicalPath())[len(portal_path):] or '/',  # noqa
                'data': self.addTilesData(effective_tile)
            })

        tile = meta.getTile(self.context, self.request, _id)
        mode = tile.data.get('mode', 'parent') or 'parent'
        if current_path == '/' and mode == 'parent':
            mode = 'show'
        return json.dumps({
            'current': current_path or '/',
            'current_url': self.context.absolute_url(),
            'mode': mode,
            'effective_tiles': effective_tiles,
            'tiles': self.addTilesData(tile)
        })


class RenderLayout(BrowserView):
    javascript = '''
$("*").each(function () {

  // Local variables
  var obj;

  obj = $(this);

  // Check if block element
  if (obj.css('display') === 'block') {
    // Add blur class
    if(!obj.hasClass('castle-tile-container')){
        obj.addClass('mosaic-blur');
    }
  }
  $('.castle-tile-container').each(function(){
    $(this).parents('.mosaic-blur').removeClass('mosaic-blur');
  });
});
'''
    styles = '''.mosaic-blur {
    opacity: 0.7;
}
.castle-tile-container {
outline: 2px dashed #ccc;
margin: 3px;
text-align: center;
font-weight: bold;
padding: 20px;
cursor: pointer;
z-index: 9999;
position: relative;
}
.castle-tile-container:hover {
outline: 2px dashed orange;
}
'''

    def __call__(self):
        self.request.response.setHeader('X-Theme-Applied', 'true')
        layout = self.request.get('layout')
        transform = theming.getTransform(self.context, self.request)

        layout = transform.get_layout(self.context, layout, request=self.request)
        portal = api.portal.get()
        portal_url = portal.absolute_url()
        context_url = self.context.absolute_url()

        theme_base_url = '%s/++%s++%s/index.html' % (
            portal_url,
            THEME_RESOURCE_NAME,
            transform.name)

        content = {
            'main': '<p>Content from page</p>',
            'left': '',
            'right': ''
        }

        utils = getMultiAdapter((self.context, self.request),
                                name='castle-utils')

        layout = layout(
            portal_url=portal_url,
            site_url=portal_url,
            context_url=context_url,
            request=self.request,
            context=self.context,
            portal=portal,
            site=portal,
            theme_base_url=theme_base_url,
            content=content,
            anonymous=api.user.is_anonymous(),
            debug=api.env.debug_mode(),
            utils=utils,
            raw=True
        )

        dom = getHTMLSerializer([layout])
        transform.rewrite(dom, theme_base_url)

        root = dom.tree.getroot()
        for tile in root.cssselect('[data-tile*="@@castle.cms.meta/"]'):
            slot_id = tile.attrib['data-tile'].split('/@@castle.cms.meta/')[-1]

            title = None
            if '?' in slot_id:
                slot_id, _, qs = slot_id.partition('?')
                query = dict(urllib.parse_qsl(qs))
                title = query.get('title')
            else:
                title = slot_id.replace('meta-', '').capitalize().replace('-', ' ')
            tile.attrib['data-tile-id'] = slot_id
            tile.attrib['data-tile-title'] = title
            tile.attrib['class'] = 'castle-tile-container'
            tile.text = 'Edit %s slot' % title

        overlay = etree.Element('div')
        overlay.attrib['class'] = 'expose-overlay'
        root.cssselect('body')[0].append(overlay)

        javascript = etree.Element('script')
        javascript.text = self.javascript
        root.cssselect('body')[0].append(javascript)

        style = etree.Element('style')
        style.text = self.styles
        root.cssselect('body')[0].append(style)

        transform.dynamic_grid(dom.tree)
        return tostring(dom.tree)



from castle.cms.interfaces import IReferenceNamedImage
from plone.app.uuid.utils import uuidToObject
from persistent.mapping import PersistentMapping
from persistent.dict import PersistentDict
from lxml.html import fromstring
from lxml.html import tostring
from plone import api
from plone.app.blocks.layoutbehavior import ILayoutAware
from plone.app.linkintegrity import handlers as li
from plone.app.linkintegrity.parser import extractLinks
from plone.tiles.data import ANNOTATIONS_KEY_PREFIX
from z3c.relationfield import RelationValue
from zope.annotation.interfaces import IAnnotations
from zope.component import getUtility
from zope.intid.interfaces import IIntIds
from zope.keyreference.interfaces import NotYet


def scan(obj):
    """ a dexterity based object was modified """
    if not li.check_linkintegrity_dependencies(obj):
        return
    refs = get_content_links(obj)
    li.updateReferences(obj, refs)


def get_ref(obj, intids=None):
    if intids is None:
        intids = getUtility(IIntIds)
    try:
        objid = intids.getId(obj)
    except KeyError:
        try:
            intids.register(obj)
            objid = intids.getId(obj)
        except NotYet:
            # if we get a NotYet error, the object is not
            # attached yet and we will need to get links
            # at a later time when the object has an intid
            pass
    return objid


def get_content_links(obj):
    refs = set()
    if ILayoutAware.providedBy(obj):
        behavior_data = ILayoutAware(obj)
        # get data from tile data
        annotations = IAnnotations(obj)
        for key in annotations.keys():
            if key.startswith(ANNOTATIONS_KEY_PREFIX):
                data = annotations[key]
                refs |= get_tile_data_links(obj, data)
        if not behavior_data.contentLayout and behavior_data.content:
            dom = fromstring(behavior_data.content)
            for el in dom.cssselect('.mosaic-text-tile .mosaic-tile-content'):
                links = extractLinks(tostring(el))
                refs |= li.getObjectsFromLinks(obj, links)
    try:
        # scan more than just this we probably should...
        value = obj.text.raw
        links = extractLinks(value)
        refs |= li.getObjectsFromLinks(obj, links)
    except AttributeError:
        pass

    if getattr(obj, 'image'):
        if IReferenceNamedImage.providedBy(obj.image):
            sub_obj = uuidToObject(obj.image.reference)
            if sub_obj:
                objid = get_ref(obj)
                if objid:
                    refs.add(RelationValue(objid))
    return refs


def get_tile_data_links(obj, data):
    refs = set()
    if type(data) in (dict, PersistentMapping, PersistentDict):
        for field_name in ('content', 'video', 'image', 'images', 'audio'):
            if field_name not in data:
                continue
            val = data.get(field_name)
            if isinstance(val, basestring):
                links = extractLinks(val)
                refs |= li.getObjectsFromLinks(obj, links)
            elif isinstance(val, list):
                # could be list of uids
                refs |= get_refs_from_uids(val)
    return refs


def get_refs_from_uids(uids):
    intids = getUtility(IIntIds)
    objects = set()
    catalog = api.portal.get_tool('portal_catalog')
    for brain in catalog(UID=uids):
        obj = brain.getObject()
        objid = get_ref(obj, intids)
        if objid:
            relation = RelationValue(objid)
            objects.add(relation)
    return objects



from Products.Five import BrowserView
from celery.task.control import inspect
from plone import api
from castle.cms import taskinfo


class CeleryControlPanel(BrowserView):

    def info(self):
        ins = inspect()
        return {
            'workers': ins.ping(),
            'active': ins.active(),
            'reserved': ins.reserved(),
            'stats': ins.stats()
        }

    def get_task_name(self, _id):
        return taskinfo.get_task_name(_id)

    def task_info(self, task):
        info = taskinfo.get_info(task)
        on_site = False
        if info['kwargs'].get('site_path') == '/'.join(self.site.getPhysicalPath()):
            on_site = True

        obj_path = ''
        obj = None
        args = info['args']
        if len(args) > 0 and isinstance(args[0], basestring):
            obj_path = args[0].replace('object://', '')
            obj = self.site.unrestrictedTraverse(str(obj_path), None)

        info.update({
            'on_site': on_site,
            'obj': obj,
            'obj_path': obj_path,
        })
        return info

    def __call__(self):
        self.site = api.portal.get()
        return self.index()
